/*
 * drivers/library/debug_misc.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */


#include <stddef.h>

#ifdef __sun__
#include <sys/types.h>
#include <sys/pci.h>
#endif

#ifdef __linux__
#include <linux/pci.h>
#endif

#include <debug_misc.h>

#define DMA_SGL_INTERRUPT	(1LL<<31)
#define DMA_GO			(1<<30)
#define DMA_FIFO_CLEAR		(1<<29)
#define DMA_ABORT_INTERRUPT	(1<<28)
#define DMA_PCI_CORE_INTERRUPT	(1<<27)
#define DMA_ECC_INTERRUPT	(1<<26)
#define DMA_UNCORRECTABLE	(1<<25)
#define DMA_CORRECTABLE		(1<<24)
#define DMA_AD3_PROTOCOL_ERROR	(1<<23)
#define DMA_INTERRUPT_PENDING	(1<<16)
#define DMA_SGL_HIGH_USED_MASK	0x0003FE00    /* 17:9 */
#define DMA_SGL_LOW_USED_MASK	0x000001FF    /* 8:0 */

#define STATLOW_MASK_NONFATAL   (1<<0)
#define STATLOW_MASK_FATAL      (1<<1)
#define HARD_DISK_FAILURE       (1<<2)
#define SDRAM_UNCORRECTABLE     (1<<3)
#define STATLOW_ALWAYS_1        (1<<6)
#define STATLOW_RD_NAK          (1<<8)
#define STATLOW_WR_NAK          (1<<9)


#define LSB(x) ((x) & -(x))


static void
strlcpy(char *dest, const char *src, size_t len)
{
    if (len == 0)
	return;
    --len;
    while (*src && len)
    {
	*dest++ = *src++;
	--len;
    }
    *dest = 0;
}


static void
strlcat(char *dest, const char *src, size_t len)
{
    if (len == 0)
	return;
    --len;
    while (*dest && len)
    {
	++dest;
	--len;
    }
    while (*src && len)
    {
	*dest++ = *src++;
	--len;
    }
    *dest = 0;
}


static const char *
dma_bit_decode(unsigned long dma_bit)
{
    switch (dma_bit)
    {
    case (unsigned long)DMA_SGL_INTERRUPT:
	return "DMA_SGL_INTERRUPT";

    case (unsigned long)DMA_GO:
        return "DMA_GO";

    case (unsigned long)DMA_FIFO_CLEAR:
        return "DMA_FIFO_CLEAR";

    case (unsigned long)DMA_ABORT_INTERRUPT:
        return "DMA_ABORT_INTERRUPT";

    case (unsigned long)DMA_PCI_CORE_INTERRUPT:
        return "DMA_PCI_CORE_INTERRUPT";

    case (unsigned long)DMA_ECC_INTERRUPT:
        return "DMA_ECC_INTERRUPT";

    case (unsigned long)DMA_UNCORRECTABLE:
        return "DMA_UNCORRECTABLE";

    case (unsigned long)DMA_CORRECTABLE:
        return "DMA_CORRECTABLE";

    case (unsigned long)DMA_AD3_PROTOCOL_ERROR:
        return "DMA_AD3_PROTOCOL_ERROR";

    case (unsigned long)DMA_INTERRUPT_PENDING:
        return "DMA_INTERRUPT_PENDING";

    default:
	return "?";
    }
}


const char *
dma_decode(unsigned long dma)
{
    static char buffer[300];
    unsigned long x;

    buffer[0] = 0;
    x = dma & ~DMA_SGL_LOW_USED_MASK;
    while (x)
    {
	unsigned long bit;

	bit = LSB(x);
	x -= bit;
	if (buffer[0])
	    strlcat(buffer, ", ", sizeof(buffer));
	strlcat(buffer, dma_bit_decode(bit), sizeof(buffer));
    }
    if (!buffer[0])
	strlcpy(buffer, "nothing", sizeof(buffer));
    return buffer;
}


static const char *
status_bit_decode(unsigned status_bit)
{
    /*
     * Unfortunately, each operating system has slightly different
     * names for each of the status bits.  Each bit is independently
     * conditionally compiled, because not all revs of all operating
     * systems define them.  Sigh.
     */
    switch (status_bit)
    {
#ifdef PCI_STAT_CAP
    case PCI_STAT_CAP:
	return "PCI_STAT_CAP";
#endif

#ifdef PCI_STAT_66MHZ
    case PCI_STAT_66MHZ:
	return "PCI_STAT_66MHZ";
#endif

#ifdef PCI_STAT_UDF
    case PCI_STAT_UDF:
	return "PCI_STAT_UDF";
#endif

#ifdef PCI_STAT_FBBC
    case PCI_STAT_FBBC:
	return "PCI_STAT_FBBC";
#endif

#ifdef PCI_STAT_S_PERROR
    case PCI_STAT_S_PERROR:
	return "PCI_STAT_S_PERROR";
#endif

#ifdef PCI_STAT_S_TARG_AB
    case PCI_STAT_S_TARG_AB:
	return "PCI_STAT_S_TARG_AB";
#endif

#ifdef PCI_STAT_R_TARG_AB
    case PCI_STAT_R_TARG_AB:
	return "PCI_STAT_R_TARG_AB";
#endif

#ifdef PCI_STAT_R_MAST_AB
    case PCI_STAT_R_MAST_AB:
	return "PCI_STAT_R_MAST_AB";
#endif

#ifdef PCI_STAT_S_SYSERR
    case PCI_STAT_S_SYSERR:
	return "PCI_STAT_S_SYSERR";
#endif

#ifdef PCI_STAT_PERROR
    case PCI_STAT_PERROR:
	return "PCI_STAT_PERROR";
#endif

#ifdef PCI_STATUS_CAP_LIST
    case PCI_STATUS_CAP_LIST:
       	return "PCI_STATUS_CAP_LIST";
#endif

#ifdef PCI_STATUS_66MHZ
    case PCI_STATUS_66MHZ:
       	return "PCI_STATUS_66MHZ";
#endif

#ifdef PCI_STATUS_UDF
    case PCI_STATUS_UDF:
       	return "PCI_STATUS_UDF";
#endif

#ifdef PCI_STATUS_FAST_BACK
    case PCI_STATUS_FAST_BACK:
       	return "PCI_STATUS_FAST_BACK";
#endif

#ifdef PCI_STATUS_PARITY
    case PCI_STATUS_PARITY:
       	return "PCI_STATUS_PARITY";
#endif

#ifdef PCI_STATUS_DEVSEL_MASK
    case PCI_STATUS_DEVSEL_MASK:
       	return "PCI_STATUS_DEVSEL_MASK";
#endif

#ifdef PCI_STATUS_DEVSEL_FAST
    case PCI_STATUS_DEVSEL_FAST:
       	return "PCI_STATUS_DEVSEL_FAST";
#endif

#ifdef PCI_STATUS_DEVSEL_MEDIUM
    case PCI_STATUS_DEVSEL_MEDIUM:
       	return "PCI_STATUS_DEVSEL_MEDIUM";
#endif

#ifdef PCI_STATUS_DEVSEL_SLOW
    case PCI_STATUS_DEVSEL_SLOW:
       	return "PCI_STATUS_DEVSEL_SLOW";
#endif

#ifdef PCI_STATUS_SIG_TARGET_ABORT
    case PCI_STATUS_SIG_TARGET_ABORT:
       	return "PCI_STATUS_SIG_TARGET_ABORT";
#endif

#ifdef PCI_STATUS_REC_TARGET_ABORT
    case PCI_STATUS_REC_TARGET_ABORT:
       	return "PCI_STATUS_REC_TARGET_ABORT";
#endif

#ifdef PCI_STATUS_REC_MASTER_ABORT
    case PCI_STATUS_REC_MASTER_ABORT:
       	return "PCI_STATUS_REC_MASTER_ABORT";
#endif

#ifdef PCI_STATUS_SIG_SYSTEM_ERROR
    case PCI_STATUS_SIG_SYSTEM_ERROR:
       	return "PCI_STATUS_SIG_SYSTEM_ERROR";
#endif

#ifdef PCI_STATUS_DETECTED_PARITY
    case PCI_STATUS_DETECTED_PARITY:
       	return "PCI_STATUS_DETECTED_PARITY";
#endif

    default:
	return "?";
    }
}


const char *
status_decode(unsigned status)
{
    static char buffer[1000];

    buffer[0] = 0;
    while (status)
    {
	unsigned bit;

	bit = LSB(status);
	status -= bit;
	if (buffer[0])
	    strlcat(buffer, ", ", sizeof(buffer));
	strlcat(buffer, status_bit_decode(bit), sizeof(buffer));
    }
    if (!buffer[0])
	strlcpy(buffer, "nothing", sizeof(buffer));
    return buffer;
}


static const char *
statlow_decode_bit(unsigned long x)
{
    switch (x)
    {
    case STATLOW_MASK_NONFATAL:
	return "SDRAM_NONFATAL_ECC";

    case STATLOW_MASK_FATAL:
	return "FATAL";

    case HARD_DISK_FAILURE:
	return "HARD_DISK_FAILURE";

    case SDRAM_UNCORRECTABLE:
	return "SDRAM_UNCORRECTABLE";

    case STATLOW_ALWAYS_1:
	return "ALWAYS_ON";

    case STATLOW_RD_NAK:
	return "RD_NAK";

    case STATLOW_WR_NAK:
	return "WR_NAK";

    default:
	return "?";
    }
}


const char *
statlow_decode(unsigned long x)
{
    static char buffer[300];

    buffer[0] = 0;
    while (x)
    {
	unsigned long bit;

	bit = LSB(x);
	x -= bit;
	if (buffer[0])
	    strlcat(buffer, ", ", sizeof(buffer));
	strlcat(buffer, statlow_decode_bit(bit), sizeof(buffer));
    }
    if (!buffer[0])
	strlcpy(buffer, "nothing", sizeof(buffer));
    return buffer;
}
