/*
 * drivers/linux2.4/fndimm.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#include <kernel_config_stuff.h>

#define __NO_VERSION__		/* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>

#include <linux/sched.h>
#include <linux/kernel.h>	/* HDEBUG() */
#include <linux/slab.h>	/* kmalloc() */
#include <linux/vmalloc.h>
#include <linux/fs.h>		/* everything... */
#include <linux/errno.h>	/* error codes */
#include <linux/timer.h>
#include <linux/types.h>	/* size_t */
#include <linux/fcntl.h>	/* O_ACCMODE */
#include <linux/hdreg.h>	/* HDIO_GETGEO */
#include <linux/pci.h>
#include <linux/param.h>

#include <asm/system.h>		/* cli(), *_flags */
#include <asm/uaccess.h>	/* put_user */
#include <asm/delay.h>
#include <asm/io.h>

#include <qikdata_hba.h>	/* hardware definitions */
#include <qikdata_drive.h>	/* local definitions */

#include <fndimm.h>
#include <qd_cable_registers.h>

typedef union _DIMM_BASIS
    {
	u32     dw;
	struct
	    {
		u8      rows;
		u8      cols;
		u8      banks;
		u8      rsrows:2;
		u8      ecc:1;
		u8      exists:1;
		u8      unused:4;
	    }
	s;
    }
DIMM_BASIS;

int
qikdata_read_dimm_spds(kdev_t i_rdev)
{
    int hba_idx = QD_HBA_NR(i_rdev);
    int drv_idx = QD_DRV_NR(i_rdev);
    PHW_DEVICE_EXTENSION devExt = & (qikdata_device_extension[hba_idx]);
    PUNIT_DETAILS unitDet = &(devExt->UnitDet[drv_idx]);
    int     dimm;
    u64     capacity;
    u32     totalBlocks = 0;
    PAGE_REG pageRegister;
    u_int   pRTR;
    u_int   pMEM;
    int     c_error;

    HDEBUG(KERN_ALERT HERE "qikdata_read_dimmm_spds()\n");
    /*
     * Set the global so I2C routines can pick it up
     */
    pRTR = (u_int) devExt->RTRmap;
    pMEM = (u_int) devExt->LASmap;
    /*
     * Assume for the moment we have all ECC memory
     */
    unitDet->ECCInstalled = 1;
    /*
     * Initialise maximum counters
     */
    unitDet->MaxDimmRows = 0;
    unitDet->MaxDimmCols = 0;
    unitDet->MaxDimmB = 0;

    pageRegister.dw = 0;
    pageRegister.s.nios = 1;
    pageRegister.s.drive = unitDet->drive_address;
    pageRegister.dw |=
	(PAGE_MASK_NIOS | PAGE_MASK_EXTENDED | PAGE_MASK_DIMM_SPACE);

    for (dimm = 0; dimm < MAX_DIMMS; dimm++)
    {
	DIMM_BASIS db;
	/*
	 * select the nios map, current dimm
	 */
	pageRegister.s.dimm = dimm;
	writel(pageRegister.dw, pRTR + REG_PAGE);
	HDEBUG(KERN_ALERT HERE "page reg: %08x\n", readl(pRTR + REG_PAGE));
	HDEBUG(KERN_ALERT HERE "spd info: %08x%08x\n",
	       readl(pMEM + CABLE_ED_SPDS),
	       readl(pMEM + CABLE_ED_SPDS + 4));

	db.dw = readl(pMEM + CABLE_ED_SPDS);
	db.dw = readl(pMEM + CABLE_ED_SPDS);
	HDEBUG(KERN_ALERT HERE "spd info : $%08x\n", db.dw);

	if (db.s.exists)
	{
	    DIMM_PARAMS *dp;

	    dp = &unitDet->Dimm[dimm];
	    dp->cols_log2 = db.s.cols;
	    dp->cols_total = 1u << db.s.cols;
	    dp->cols_mask = dp->cols_total - 1;
	    dp->rows_log2 = db.s.rows;
	    dp->rows_total = 1u << db.s.rows;
	    dp->rows_mask = dp->rows_total - 1;
	    /*
	     * db.s.rsrows can only assume values of 1 and 2.
	     * db.s.banks can only assume values of 1, 2 and 4.
	     *
	     * The trick is: log2(n) ~= n/2, for the values we are
	     * intested in, e.g.
	     *     log2(4) = 2
	     *     log2(2) = 1
	     *     log2(1) = 0
	     */
	    dp->banks_log2 = db.s.banks / 2;
	    dp->banks_total = db.s.banks;
	    dp->banks_mask = dp->banks_total - 1;
	    dp->rsrows_log2 = db.s.rsrows / 2;
	    dp->rsrows_total = db.s.rsrows;
	    dp->rsrows_mask = dp->rsrows_total - 1;

	    dp->ecc = db.s.ecc;
	    dp->assembly_serial = readl(pMEM + CABLE_ED_SPDS + 4);
	    c_error = 0;
	    udelay(5);
	}
	else
	{
	    DIMM_PARAMS *dp;

	    /*
	     * dimm is missing
	     */
	    dp = &unitDet->Dimm[dimm];
	    dp->cols_log2 = 0;
	    dp->cols_total = 0;
	    dp->cols_mask = 0;
	    dp->rows_log2 = 0;
	    dp->rows_total = 0;
	    dp->rows_mask = 0;
	    dp->banks_log2 = 0;
	    dp->banks_total = 0;
	    dp->banks_mask = 0;
	    dp->rsrows_log2 = 0;
	    dp->rsrows_total = 0;
	    dp->rsrows_mask = 0;
	    dp->ecc = 0;
	    c_error = 1;
	}
	HDEBUG(KERN_ALERT HERE "DIMM %d\n", dimm);

	if (!c_error)
	{
	    HDEBUG(KERN_ALERT HERE
		"Manufacturer = \"%-8.8s\", Assembly Serial = %u\n",
		unitDet->Dimm[dimm].manufacturer_jedec_id,
		unitDet->Dimm[dimm].assembly_serial);
	    HDEBUG(KERN_ALERT HERE
		"RsRows = %d (%d bits), Banks = %d (%d bits)\n",
		unitDet->Dimm[dimm].rsrows_total,
		unitDet->Dimm[dimm].rsrows_log2,
		unitDet->Dimm[dimm].banks_total,
		unitDet->Dimm[dimm].banks_log2);
	    HDEBUG(KERN_ALERT HERE
		"Rows = %d (%d bits), Cols = %d (%d bits)\n",
		unitDet->Dimm[dimm].rows_total,
		unitDet->Dimm[dimm].rows_log2,
		unitDet->Dimm[dimm].cols_total,
		unitDet->Dimm[dimm].cols_log2);

	    /*
	     * Store the maximum rows, cols, b for sdram init optimisation
	     */
	    if (unitDet->Dimm[dimm].rows_log2 > unitDet->MaxDimmRows)
		unitDet->MaxDimmRows = unitDet->Dimm[dimm].rows_log2;
	    if (unitDet->Dimm[dimm].cols_log2 > unitDet->MaxDimmCols)
		unitDet->MaxDimmCols = unitDet->Dimm[dimm].cols_log2;
	    if (unitDet->Dimm[dimm].banks_log2 > unitDet->MaxDimmB)
		unitDet->MaxDimmB = unitDet->Dimm[dimm].banks_log2;

	    capacity = qikdata_dimm_capacity(&unitDet->Dimm[dimm]);

#if 0
	    HDEBUG(KERN_ALERT HERE "Capacity %d%06d bytes, %d MB\n",
		(int)(capacity / 1000000), (int)(capacity % 1000000),
		(int)(capacity / 1000000));
#endif

	    /*
	     * If we find a non-ECC dimm, turn off ECC functionality
	     */
	    if (unitDet->Dimm[dimm].ecc != 0x02)
		unitDet->ECCInstalled = 0;
	}
	else
	{
	    /*
	     * set serial number to indicate not present
	     */
	    unitDet->Dimm[dimm].assembly_serial = -1;

	    HDEBUG(KERN_ALERT HERE "Not found!\n");

	    capacity = 0;
	}
	/*
	 * initialise some useful values/masks for optimisation
	 */
	unitDet->Dimm[dimm].startBlock = totalBlocks;
	unitDet->Dimm[dimm].blocks = (u32) capacity >> 9;
	unitDet->Dimm[dimm].totalBlocks =
	    unitDet->Dimm[dimm].startBlock + unitDet->Dimm[dimm].blocks;
	totalBlocks += unitDet->Dimm[dimm].blocks;
	unitDet->Dimm[dimm].baseAddress = dimm << 29;

        HDEBUG(KERN_ALERT HERE
           "startBlock = %d, blocks = %d, totalBlocks = %d, base=$"
           HEX8 "\n", unitDet->Dimm[dimm].startBlock,
           unitDet->Dimm[dimm].blocks, unitDet->Dimm[dimm].totalBlocks,
           unitDet->Dimm[dimm].baseAddress);
    }
    /*
     * restore the page register
     */
    writel(0x00, pRTR + REG_PAGE);

    HDEBUG(KERN_ALERT HERE "Total drive capacity = %u bytes (%uMB)\n",
	totalBlocks << 9, totalBlocks >> 11);
    HDEBUG(KERN_ALERT HERE "ECC memory is %sinstalled\n",
	(unitDet->ECCInstalled ? "" : "not "));
    HDEBUG(KERN_ALERT HERE "Maximum bits Rows %d, Cols %d, Banks %d\n",
	unitDet->MaxDimmRows, unitDet->MaxDimmCols, unitDet->MaxDimmB);
    /*
     * Calculate the size of the sdram initialisation
     */
    devExt->SdramInitBlitBlocks =
	1 << (3 + unitDet->MaxDimmRows + unitDet->MaxDimmCols +
	unitDet->MaxDimmB - 9);

    HDEBUG(KERN_ALERT HERE "SdramInitBlitBlocks = %d\n",
	devExt->SdramInitBlitBlocks);
    /*
     * Assume we have installed memory in the drive
     */
    /*
     * ASSERT( devExt->ScratchpadBlockAddress != 0 );
     */
#ifdef _NEVER_ENABLE_ECC
    unitDet->ECCInstalled = 0;
#endif

    return ((u8) 0);
}

/*
 * Returns physical address of requested block
 * convert contiguous address to LB address
 */
u_long
qikdata_addr2lb(int drive_address, PDIMM_PARAMS dimm, u64 blockAddress,
    PSGL_ENTRY pSgl)
{
    u64     addr;		       /* linear dimm offset address */
    u64     lbaddr;		       /* target address */
    int     d;			       /* dimm number */

    /*
     * initialise banking register
     */
    pSgl->qw = (u64) 0;

    /*
     * pSgl->s.drive is passes from devExt->drive_address
     */
    pSgl->s.drive = drive_address;
    /*
     * Calculate the dimm and address offset
     *
     * That 9 is log2(hardsectsize)
     */
    for (d = 0; d < MAX_DIMMS; d++)
	if ((blockAddress >> 9) < dimm[d].totalBlocks)
	    break;
    /*
     * attempt to read past end of device returns error
     */
    if (d == MAX_DIMMS)
	return ((u_long) (-1L));
    /*
     * adjust pointer to actual dimm
     */
    dimm += d;
    pSgl->s.dimm = d;
    /*
     * calculate linear dimm offset address
     *
     * That 9 is log2(hardsectsize)
     */
    addr = blockAddress - (dimm->startBlock << 9);
    /*
     * put in byte addr
     */
    lbaddr = addr & 7;
    addr >>= 3;
    /*
     * put in cols
     */
    pSgl->s.col = addr & dimm->cols_mask;
    addr >>= dimm->cols_log2;
    /*
     * rows
     */
    pSgl->s.row = addr & dimm->rows_mask;
    addr >>=  dimm->rows_log2;
    /*
     * banks
     */
    pSgl->s.bank = addr & dimm->banks_mask;
    addr >>= dimm->banks_log2;
    /*
     * rs rows
     */
    pSgl->s.rsrows = addr & dimm->rsrows_mask;

#ifdef QIKDATA_DEBUG
    addr >>= dimm->rsrows_log2;
    if (addr)
	HDEBUG(KERN_ALERT HERE "addr2bl overflow: %d\n", (int)addr);
#endif

    return (lbaddr);
}

u64
qikdata_dimm_capacity(PDIMM_PARAMS dimm)
{
    return (1UL << (dimm->rows_log2 + dimm->cols_log2 + dimm->banks_log2 +
	dimm->rsrows_log2 + 3));
}

void
qikdata_initialize_dimm_statictics(kdev_t i_rdev)
{
    int hba_idx = QD_HBA_NR(i_rdev);
    int drv_idx = QD_DRV_NR(i_rdev);
    PHW_DEVICE_EXTENSION devExt = & (qikdata_device_extension[hba_idx]);
    PUNIT_DETAILS unitDet = &(devExt->UnitDet[drv_idx]);
    int     i;

    for (i = 0; i < MAX_DIMMS; i++)
    {
	unitDet->DimmStats[i].read_blocks = 0;
	unitDet->DimmStats[i].write_blocks = 0;
	unitDet->DimmStats[i].correctable = 0;
	unitDet->DimmStats[i].uncorrectable = 0;
    }
}
