/*
 * drivers/linux2.4/fnioctl.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#include <kernel_config_stuff.h>

#define __NO_VERSION__		/* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>	/* printk() */
#include <linux/slab.h>	/* kmalloc() */
#include <linux/vmalloc.h>
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/ioctl.h>
#include <asm/uaccess.h>	/* put_user */
#include <asm/delay.h>
#include <asm/io.h>

#include <qikdata_hba.h>	/* hardware definitions */
#include <qikdata_drive.h>	/* local definitions */
#include <fndimm.h>

#include <fnexec.h>
#include <fnioctl.h>

static const char BLOCK_TEST[] = "bt ";
static const char GOOD[] = "good ";
static const char FAULT[] = "flt ";
static const char OVRD[] = "ovrd ";
static const char SPD[] = "spd ";
static const char STATS[] = "stats ";
static const char RMEM[] = "rmem ";
static const char INT[] = "int ";
static const char LRESET[] = "lreset ";

/* a 16k buffer for output to copy back to user land */
static char buffer[16384];

static void Handle_BlockTest(kdev_t i_rdev, char *buffer);
static void Handle_GoodLed(kdev_t i_rdev, char *buffer);
static void Handle_FaultLed(kdev_t i_rdev, char *buffer);
static void Handle_Override(kdev_t i_rdev, char *buffer);
static void Handle_ForceReadDimmSpds(kdev_t i_rdev, char *buffer);
static void Handle_Statistics(kdev_t i_rdev, char *buffer);
static void Handle_ReadMem(kdev_t i_rdev, char *buffer);
static void Handle_MemReadWriteBlock(kdev_t i_rdev,
    long int block, int write, char *buffer);
static void Handle_Interrupt(kdev_t i_rdev, char *buffer);

static void
usage(char *buffer)
{
    int     len = 0;

    len += sprintf(buffer + len, "Commands:\n");
    len +=
	sprintf(buffer + len, "  %-32.32s : Block Test (memory i/o)\n",
	"bt [<num> [<start>]]");
    len +=
	sprintf(buffer + len, "  %-32.32s : GOOD LED control\n", "good {0|1}");
    len +=
	sprintf(buffer + len, "  %-32.32s : FAULT LED control\n", "flt {0|1}");
    len +=
	sprintf(buffer + len, "  %-32.32s : Set indicator override bits\n",
	"ovrd [bitmask]");
    len +=
	sprintf(buffer + len, "  %-32.32s : Force read of DIMM SPDs\n", "spd");
    len +=
	sprintf(buffer + len, "  %-32.32s : [Clear] firmware statistics\n",
	"stats [c]");
    len +=
	sprintf(buffer + len, "  %-32.32s : Read block <n> from device\n",
	"rmem <n>");
    len +=
	sprintf(buffer + len, "  %-32.32s : Start dummy DMA to hit interrupt\n",
	"int");
    len +=
	sprintf(buffer + len, "  %-32.32s : Complete Outstanding Request\n",
	"cr");
    len += sprintf(buffer + len, "  %-32.32s : Force Link Reset\n", "lreset");
}

int
qikdata_handle_ioctl(kdev_t i_rdev, u_int cmd, u_long arg)
{
    char   *p = buffer;

    /*
     * get the arguments
     */
    if (copy_from_user(buffer, (char *)arg, 16384) > 0)
    {
	HDEBUG(KERN_ALERT HERE "unable to copy from user space!\n");
	return (-1);
    }

    if (!strncmp(p, BLOCK_TEST, strlen(BLOCK_TEST)))
    {
	Handle_BlockTest(i_rdev, p);
    }
    else if (!strncmp(p, GOOD, strlen(GOOD)))
    {
	Handle_GoodLed(i_rdev, p);
    }
    else if (!strncmp(p, FAULT, strlen(FAULT)))
    {
	Handle_FaultLed(i_rdev, p);
    }
    else if (!strncmp(p, OVRD, strlen(OVRD)))
    {
	Handle_Override(i_rdev, p);
    }
    else if (!strncmp(p, SPD, strlen(SPD)))
    {
	Handle_ForceReadDimmSpds(i_rdev, p);
    }
    else if (!strncmp(p, STATS, strlen(STATS)))
    {
	Handle_Statistics(i_rdev, p);
    }
    else if (!strncmp(p, RMEM, strlen(RMEM)))
    {
	Handle_ReadMem(i_rdev, p);
    }
    else if (!strncmp(p, INT, strlen(INT)))
    {
	Handle_Interrupt(i_rdev, p);
    }
    else if (!strncmp(p, LRESET, strlen(LRESET)))
    {
	HDEBUG(KERN_ALERT HERE "Sending link reset!\n");
	/*
	 * send a link reset
	 */
	qikdata_link_reset(i_rdev);
    }
    else
    {
	/*
	 * knock off the trailing space
	 */
	p[strlen(p) - 1] = '\0';
	usage(p);
    }

    /*
     * copy back to user land
     */
    if (copy_to_user((char *)arg, buffer, 16384) > 0)
    {
	HDEBUG(KERN_ALERT HERE "unable to copy buffer to user space!\n");
	return (-1);
    }

    return (0);
}

static int
isspace(char p)
{
    return (p == ' ' || p == '\t');
}

static int
isdigit(char p)
{
    return (p >= '0' && p <= '9');
}


static long int
atoi(char *p)
{
    long int value = 0;
    long int sign = 1;

    if (*p == '-')
    {
	sign = -sign;
	p++;
    }

    while (*p && isdigit(*p))
    {
	value = value * 10 + *p - '0';
	p++;
    }
    return (sign * value);
}

static int DumpBlock(char *buffer, char *memBuf);

static void
Handle_BlockTest(kdev_t i_rdev, char *buffer)
{
    PHW_DEVICE_EXTENSION devExt = &qikdata_device_extension[QD_HBA_NR(i_rdev)];
    static unsigned char memBuf[512];
    PUNIT_DETAILS unitDet = & (devExt->UnitDet[QD_DRV_NR(i_rdev)]);

    int     startBlock = 0,
            nBlocks = 0;
    int     i;
    int     len = 0;
    char   *p = buffer;

    p += strlen(BLOCK_TEST);
    while (*p && isspace(*p))
	p++;
    if (!(*p))
	nBlocks = unitDet->Dimm[MAX_DIMMS - 1].totalBlocks;
    else
	nBlocks = atoi(p);

    while (*p && !isspace(*p))
	p++;
    while (*p && isspace(*p))
	p++;
    if (!(*p))
	startBlock = 0;
    else
	startBlock = atoi(p);

    /*
     * Check the bounds of the test
     */
    if (startBlock + nBlocks > unitDet->Dimm[MAX_DIMMS - 1].totalBlocks)
	nBlocks = unitDet->Dimm[MAX_DIMMS - 1].totalBlocks - startBlock;

    len +=
	sprintf(buffer + len, "Block Test: Start=%d, nBlocks=%d\n", startBlock,
	nBlocks);

    for (i = startBlock; i < startBlock + nBlocks; i++)
    {

	int     j;
	u_long *lbuf = (u_long *) memBuf;

	for (j = 0; j < 512 >> 2; j++)
	{
	    lbuf[j] = (i << 9) | j;
	}

	Handle_MemReadWriteBlock(i_rdev, i, 1, memBuf);
    }

    for (i = startBlock; i < startBlock + nBlocks; i++)
    {
	int     j;
	u_long  *lbuf = (u_long *)memBuf;

	Handle_MemReadWriteBlock(i_rdev, i, 0, memBuf);

	for (j = 0; j < 512 >> 2; j++)
	{
	    if (lbuf[j] != ((i << 9) | j))
	    {
		len +=
		    sprintf(buffer + len, "error: block=%d, dword@$%04X, "
			"wrote=$%08X, read=$%08lX\n", i, j << 2, (i << 9) | j,
			lbuf[j]);
		len += DumpBlock(buffer + len, memBuf);
		break;
	    }
	}

	/*
	 * did we catch an error in the last block?
	 */
	if (j != 512 >> 2)
	    break;

    }

    len += sprintf(buffer + len, "Done!\n");
}

static int
isprint(unsigned char p)
{
    return (p >= 0x20 && p <= 0x7F);
}

static int
DumpBlock(char *buffer, char *memBuf)
{
    int     len = 0;
    int     l = 0;

    for (l = 0; l < 32; l++)
    {
	int     j;

	len += sprintf(buffer + len, "%08X ", l << 4);
	for (j = 0; j < 16; j++)
	{
	    len += sprintf(buffer + len, "%02X", (int)(unsigned char)memBuf[j]);
	    if (j == 7)
		len += sprintf(buffer + len, "-");
	    else
		len += sprintf(buffer + len, " ");
	}
	len += sprintf(buffer + len, "  ");

	for (j = 0; j < 16; j++)
	{
	    if (isprint(memBuf[j]))
		len += sprintf(buffer + len, "%c", memBuf[j]);
	    else
		len += sprintf(buffer + len, ".");
	}

	len += sprintf(buffer + len, "\n");

	memBuf += 16;
    }

    return (len);
}

static void
Handle_GoodLed(kdev_t i_rdev, char *buffer)
{
}

static void
Handle_FaultLed(kdev_t i_rdev, char *buffer)
{
}

static void
Handle_Override(kdev_t i_rdev, char *buffer)
{
}

static void
Handle_ForceReadDimmSpds(kdev_t i_rdev, char *buffer)
{
    int     len = 0;

    qikdata_read_dimm_spds(i_rdev);

    len += sprintf(buffer + len, "Done!\n");
}

static void
Handle_Statistics(kdev_t i_rdev, char *buffer)
{
}


static void
Handle_ReadMem(kdev_t i_rdev, char *buffer)
{
    static unsigned char memBuf[512];
    char    *p = buffer;
    int     block = 0;

    p += strlen(RMEM);
    while (*p && isspace(*p))
	p++;

    if (*p)
    {
	block = atoi(p);
    }

    Handle_MemReadWriteBlock(i_rdev, block, 0, memBuf);
    DumpBlock(buffer, memBuf);
}

static void
Handle_MemReadWriteBlock(kdev_t i_rdev, long int block, int write,
    char *buffer)
{
}

#define nTEST_READ

static void
Handle_Interrupt(kdev_t i_rdev, char *buffer)
{
}
