/*
 * drivers/linux2.4/fnproc.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#include <kernel_config_stuff.h>

#define __NO_VERSION__		/* don't define kernel_version in module.h */
#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>	/* printk() */
#include <linux/vmalloc.h>
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */
#include <linux/proc_fs.h>
#include <linux/pci.h>

#include <asm/uaccess.h>	/* put_user */
#include <asm/io.h>

#include <qikdata_hba.h>	/* qd hardware definitions */
#include <qikdata_drive.h>	/* local definitions */
#include <fndimm.h>

#include <fnproc.h>
#include <qd_version.h>

#define LIMIT (PAGE_SIZE-80)	/* don't print any more after this size */

#ifdef LINUX_24
static struct proc_dir_entry *qd_proc_entry;
#else
static struct proc_dir_entry qd_proc_entry[MAX_CARDS];
static char proc_entry_name[MAX_CARDS][16];
#endif

#ifdef LINUX_24
static int qd_read_procmem(char *buf, char **start, off_t offset, int len,
    int *eof, void *data);
#else
static int qd_read_procmem(char *buf, char **start, off_t offset, int len,
    int deviceIndex);

static int qd_read_procmem0(char *buf, char **start, off_t offset, int len,
    int unused);

static int qd_read_procmem1(char *buf, char **start, off_t offset, int len,
    int unused);

static int qd_read_procmem2(char *buf, char **start, off_t offset, int len,
    int unused);

static int qd_read_procmem3(char *buf, char **start, off_t offset, int len,
    int unused);

static int qd_read_procmem4(char *buf, char **start, off_t offset, int len,
    int unused);

static int qd_read_procmem5(char *buf, char **start, off_t offset, int len,
    int unused);

static int qd_read_procmem6(char *buf, char **start, off_t offset, int len,
    int unused);

static int qd_read_procmem7(char *buf, char **start, off_t offset, int len,
    int unused);

static int (*read_procs[MAX_CARDS])(char *, char **, off_t, int, int) =
{
    &qd_read_procmem0,
    &qd_read_procmem1,
    &qd_read_procmem2,
    &qd_read_procmem3,
    &qd_read_procmem4,
    &qd_read_procmem5,
    &qd_read_procmem6,
    &qd_read_procmem7
};

static void
RegisterProc(PHW_DEVICE_EXTENSION devExt)
{
    int     i = devExt->DeviceIndex;

    /*
     * Initialise proc directory entry
     */
    qd_proc_entry[i].low_ino = 0;
    sprintf(proc_entry_name[i], "%s%c", "qb", 'a' + i);
    qd_proc_entry[i].namelen = strlen(proc_entry_name[i]);
    qd_proc_entry[i].name = proc_entry_name[i];
    qd_proc_entry[i].mode = S_IFREG | S_IRUGO;
    qd_proc_entry[i].nlink = 1;
    qd_proc_entry[i].uid = 0;
    qd_proc_entry[i].gid = 0;
    qd_proc_entry[i].size = 0L;
    qd_proc_entry[i].ops = NULL;
    qd_proc_entry[i].get_info = read_procs[i];

    /*
     * register the /proc filesystem handler
     */
    proc_register(&proc_root, &qd_proc_entry[i]);
}

static void
UnregisterProc(PHW_DEVICE_EXTENSION devExt)
{
    int     i = devExt->DeviceIndex;

    /*
     * unregister /proc filesystem handler
     */
    proc_unregister(&proc_root, qd_proc_entry[i].low_ino);
}

static int
qd_read_procmem0(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 0));
}

static int
qd_read_procmem1(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 1));
}

static int
qd_read_procmem2(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 2));
}

static int
qd_read_procmem3(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 4));
}

static int
qd_read_procmem4(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 4));
}

static int
qd_read_procmem5(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 5));
}

static int
qd_read_procmem6(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 6));
}

static int
qd_read_procmem7(char *buf, char **start, off_t offset, int len, int unused)
{
    return (qd_read_procmem(buf, start, offset, len, 7));
}

#endif

void
qikdata_proc_register(kdev_t i_rdev)
{
#if CONFIG_PROC_FS
    PHW_DEVICE_EXTENSION devExt = &qikdata_device_extension[QD_HBA_NR(i_rdev)];

    HDEBUG(KERN_ALERT HERE "register the /proc filesystem handler.\n");
#ifdef LINUX_24
    if (qd_proc_entry == NULL)
    {
	qd_proc_entry = proc_mkdir("driver/" DRIVER_LONG_NAME, NULL);
	if (!qd_proc_entry)
	    return;
    }
    create_proc_read_entry(devExt->qd_name, 0, qd_proc_entry, qd_read_procmem,
	(void *) ((ulong) i_rdev) );
#else
/* here for pre 2.4 kernels */
    RegisterProc(devExt);
#endif
#endif
}

void
qikdata_proc_unregister(kdev_t i_rdev)
{
#if CONFIG_PROC_FS
    PHW_DEVICE_EXTENSION devExt = &qikdata_device_extension[QD_HBA_NR(i_rdev)];
    /*
     * unregister /proc filesystem handler
     */
#ifdef LINUX_24
    if (qd_proc_entry)
	remove_proc_entry(devExt->qd_name, qd_proc_entry);
#else
    /* here for non 2.4 kernels */
    UnregisterProc(devExt);
#endif
#endif
}

void
qikdata_proc_unregister_dir()
{
#if CONFIG_PROC_FS
#ifdef LINUX_24
    if (qd_proc_entry)
    {
	HDEBUG(KERN_ALERT HERE "remove_proc_entry\n");
	remove_proc_entry(DRIVER_LONG_NAME, proc_root_driver);
	qd_proc_entry = 0;
    }
#else
    /* here for non 2.4 kernels */
#endif
#endif
}

static int
XF(int type, int dimm)
{
    if ((type & 0x000f) < 4)
	return (dimm + 1);
    else
	return (type - dimm);
}

static int
#ifdef 	LINUX_24
qd_read_procmem(char *buf, char **start, off_t offset, int len, int *eof,
    void *data)
#else
qd_read_procmem(char *buf, char **start, off_t offset, int len, int deviceIndex)
#endif
{
#ifdef LINUX_24
    kdev_t  i_rdev = (kdev_t) ((ulong) data);
    int hba_idx = QD_HBA_NR(i_rdev);
    int drv_idx = QD_DRV_NR(i_rdev);
    PHW_DEVICE_EXTENSION devExt = & (qikdata_device_extension[hba_idx]);
#else
    int hba_idx = deviceIndex;
    PHW_DEVICE_EXTENSION devExt = & (qikdata_device_extension[hba_idx]);
    int drv_idx = devExt->UnitIndex;  /* for 2.2 compatibility */
    kdev_t  i_rdev = MKDEV(devExt->major, QD_MK_MIN(hba_idx, drv_idx, 0));
#endif
    PUNIT_DETAILS unitDet;
    u_int   pRTR = (u_int) devExt->RTRmap;
    int     type = devExt->QikDriveType & 0x000F;
    int     i,j;
    u_long  time = devExt->StartTime.tv_sec;
    int     dd, hh, mm, ss;
    u_long  ecc;

    len = 0;

    /*
     * need to do this properly
     */
    dd = time / (24 * 60 * 60);
    time -= dd * (24 * 60 * 60);
    hh = time / (60 * 60);
    time -= hh * (60 * 60);
    mm = (int)(time / 60);
    time -= mm * 60;
    ss = (int)(time);

    /*
     * WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
     *
     * If you change any of this text, you must make absolutely sure
     * you don't break the code in programs/qd_qms/main.c
     */
    len += sprintf(buf + len,
	"  Logging : proc filesystem\n");

    len += sprintf(buf + len,
	"Installed : #%1d PLATYPUS QikDATA Serial#:%010u\n",
	devExt->DeviceIndex, devExt->SerialNumber);

    len += sprintf(buf + len, " Firmware : Version %04X\n", devExt->AcexIDReg);
    len += sprintf(buf + len, "   Driver : Linux Driver %s\n", qd_version());

    unitDet = & (devExt->UnitDet[drv_idx]);
    len += sprintf(buf + len,
	"\nQikDATA drives found = %d, current hba[%d].drive[%d].addr = %d\n",
	devExt->nBays, hba_idx, drv_idx, unitDet->drive_address);

    for (j = 0;j < devExt->nBays; j++)
    {
	unitDet = & (devExt->UnitDet[j]);
	len +=
	    sprintf
	    (
		buf + len,
		"\nhba[%d].drive[%d].addr = %d\n",
		hba_idx,
		j,
		unitDet->drive_address
	    );

	for (i = MAX_DIMMS - 1; i >= 0; i--)
	{
	    if (unitDet->Dimm[i].assembly_serial == -1)
		continue;

	    len +=
		sprintf
		(
		    buf + len,
		    " Dimm XF%1d : %4dMB %-3.3s %-8.8s Serial#:%010u\n",
		    XF(type, i),
		    (int)(qikdata_dimm_capacity(&(unitDet->Dimm[i])) >> 20),
		    (unitDet->Dimm[i].ecc == 0x02 ? "ECC" : ""),
		    unitDet->Dimm[i].manufacturer_jedec_id,
		    unitDet->Dimm[i].assembly_serial
		);
	}

	len += sprintf(buf + len, "Dimm Statistics :\n");
	len += sprintf(buf + len, "     %14s %14s %14s %14s\n", "Read",
	    "Write", "Uncorrectable", "Correctable");
	len += sprintf(buf + len, "      %14s %14s %14s %14s\n", "Blocks",
	    "Blocks", "Errors", "Errors");

	spin_lock_irq(&devExt->lock);
	for (i = MAX_DIMMS - 1; i >= 0; i--)
	{
	    if (unitDet->Dimm[i].assembly_serial == -1)
		continue;
	    len += sprintf(buf + len,
		"XF%1d : %14u %14u %14u %14u\n", XF(type, i),
		unitDet->DimmStats[i].read_blocks,
		unitDet->DimmStats[i].write_blocks,
		unitDet->DimmStats[i].uncorrectable,
		unitDet->DimmStats[i].correctable);
	}
	spin_unlock_irq(&devExt->lock);
    }

#ifdef QIKDATA_DEBUG
    len += sprintf(buf + len,
	"%s = %s\n", "DeviceReady",
	(devExt->DeviceReady ? "Yes" : "No"));
    len += sprintf(buf + len,
	"%s = %d\n", "State", devExt->State);
#endif

    /*
     * len += sprintf( buf+len,
     */
    len += sprintf(buf + len,
	"\nPartition Tables:");
    i_rdev = MKDEV(devExt->major, QD_MK_MIN(hba_idx, drv_idx, 0));
    len += qikdata_dump_partition_table(i_rdev, buf + len);
    len += sprintf(buf + len,
	"\nDriver Statistics:\n");
    len += sprintf(buf + len,
	"%32.32s = %ld\n", "Largest Transfer Size",
	devExt->LargestTransferSize * 512);
    len += sprintf(buf + len,
	"%32.32s = %d\n", "Received Unaligned Transfer",
	devExt->ReceivedUnaligned);
    ecc = readl(pRTR + REG_ECC);
    writel(0, pRTR + REG_ECC);	/* clear ecc */
    len += sprintf(buf + len,
	"%32.32s = %lx\n", "ECC Counts", ecc);
#ifdef QIKDATA_DEBUG
    len += sprintf(buf + len,
	"[%4d bytes written to /proc]\n", len + 32);
#endif

    /*
     * Clear statistics for next reporting period
     */
    qikdata_initialize_dimm_statictics(i_rdev);
#ifdef LINUX_24
    *eof = 1;
#endif
    return (len);
}
