/*
 * drivers/linux2.4/main.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#include <kernel_config_stuff.h>

#include <linux/module.h>
#include <linux/version.h>

#include <linux/sched.h>
#include <qd_ioctl.h>
#include <linux/wrapper.h>
#include <linux/kernel.h>	/* HDEBUG() */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/vmalloc.h>
#include <linux/fs.h>		/* everything... */
#include <linux/proc_fs.h>
#include <linux/errno.h>	/* error codes */
#include <linux/timer.h>
#include <linux/types.h>	/* size_t */
#include <linux/fcntl.h>	/* O_ACCMODE */
#include <linux/hdreg.h>	/* HDIO_GETGEO */
#include <linux/pci.h>
#include <linux/param.h>

#include <asm/system.h>		/* cli(), *_flags */
#include <asm/uaccess.h>	/* put_user */
#include <asm/delay.h>
#include <asm/irq.h>
#include <asm/io.h>
/* #include <sysdep.h> */
#include <qd_cable_registers.h>
#include <debug_misc.h>

#define MAJOR_NR qikdata_major	/* force definitions on in blk.h */
int     qikdata_major;	        /* must be declared before including blk.h */

#define DEVICE_NAME DRIVER_LONG_NAME /* name for messaging */
#define DEVICE_INTR qd_intrptr	/* pointer to the bottom half */
#define DEVICE_NO_RANDOM	/* no entropy to contribute */
#define DEVICE_REQUEST qikdata_request
#define DEVICE_ON(d)		/* do-nothing */
#define DEVICE_OFF(d)		/* do-nothing */

#define QD_SHIFT                3
#define DEVICE_NR(device)       (MINOR(device)>>QD_SHIFT)
#define PART_NR(minor)          ((minor)&((1<<QD_SHIFT)-1))
#define MINOR_NR(dev,part)      ((dev<<QD_SHIFT)+part)

#define ISR_DEBUG               KERN_ALERT
#define MAX_TRIES               10

#include <plx9054.h>		/* PCI definitions */
#include <qikdata_hba.h>	/* qd hardware definitions */
#include <qikdata_drive.h>	/* local definitions */
#include <fnproc.h>		/* /proc filesystem routines */
#include <fndimm.h>		/* dimm spd & geometry calculations */
#include <fnexec.h>		/* block device command execution */
#include <fnioctl.h>		/* io control functions */
#include <fndiag.h>		/* diagnostic functions */

#include <linux/blk.h>
#ifdef HAVE_BLKPG_H
#include <linux/blkpg.h>	/* blk_ioctl() */
#endif

#include <qd_version.h>

/*
 * We use the following fake values when we are asked for out geometry,
 * and calculate our fake cylinders based on them.
 */
#define FAKE_HEADS 255
/* #define FAKE_SECTORS 63 */
#define FAKE_SECTORS 32

/*
 * Non-prefixed symbols are static. They are meant to be assigned at
 * load time. Prefixed symbols are not static, so they can be used in
 * debugging. They are hidden anyways by register_symtab() unless
 * QIKDATA_DEBUG is defined.
 */

static int devs = QIKDRIVE_DEVS;
static int rahead = QIKDRIVE_RAHEAD;
static int blksize = QIKDRIVE_BLKSIZE;
static int hardsect = QIKDRIVE_HARDSECT;

int     qikdata_cable_ecc = 1;
int     qikdata_hardsect;
static int qd_devs;
static int qd_rahead;
static int qd_blksize;
static volatile int target_abort;

/* declared static instead of being kmalloc'd (ho m'ho ??) */
static struct hd_struct qd_partitions[MINOR_NR(QD_MAXRDEV, 0)];

/*
 * qd_sizes holds the size of device, in 1KB blocks,
 * indexed by minor device number.
 */
static int qd_sizes[MINOR_NR(QD_MAXRDEV, 0)];

static int qd_blocksizes[MINOR_NR(QD_MAXRDEV, 0)];
static int qd_hardsects[MINOR_NR(QD_MAXRDEV, 0)];

/* Driver configuration #defines... */

HW_DEVICE_EXTENSION qikdata_device_extension[MAX_CARDS];

#define _INSERT_GENDISK_IN_LIST
static char dump_buff[4000];

/*  Local function prototypes */
static int qikdata_revalidate(kdev_t i_rdev);

#ifdef LINUX_24
static struct block_device_operations qikdata_fops;
#endif

/* Our generic hard disk */

static struct gendisk qikdata_gendisk =
{
    major:0,
    major_name:DRIVER_LONG_NAME,
    minor_shift:QD_SHIFT,
    max_p:1 << QD_SHIFT,
#ifdef LINUX_24
    fops:&qikdata_fops
#else
    max_nr:QD_MAXRDEV,
    part:qd_partitions,
    sizes:qd_sizes,
    nr_real:0
#endif
};


/* Open and close */

static int
qikdata_open(struct inode *inode, struct file *filp)
{
    int     hba_idx = QD_HBA_NR(inode->i_rdev);
    PHW_DEVICE_EXTENSION devExt = &(qikdata_device_extension[hba_idx]);
    int     iflags;

    HDEBUG(KERN_WARNING HERE "opening device [%d:%d], QikDATA HBA %d, "
	"QikDATA drive %d, drive front panel address %d\n",
	MAJOR(inode->i_rdev), MINOR(inode->i_rdev),
	hba_idx, QD_DRV_NR(inode->i_rdev),
	devExt->UnitDet[QD_DRV_NR(inode->i_rdev)].drive_address);

    if (hba_idx >= qd_devs)
    {
	HDEBUG(KERN_ALERT HERE "hba_idx >= qd_devs\n");
	return (-ENODEV);
    }
    /*
     * return immediately if device is busy and non-blocking call
     */
    if ((filp->f_flags & O_NONBLOCK) && !devExt->DeviceReady)
	return (-EAGAIN);

    MOD_INC_USE_COUNT;

    /*
     * Make sure the device is ready
     * - probably don't need this here anymore since the driver
     * blocks until the hardware initialisation is complete...
     */
    iflags = 0;
    HDEBUG(KERN_ALERT HERE "spin_lock_irqsave(&devExt->lock);\n");
    spin_lock_irqsave(&devExt->lock, iflags);
    devExt->usage++;
    HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&devExt->lock);\n");
    spin_unlock_irqrestore(&devExt->lock, iflags);

    HDEBUG(KERN_ALERT HERE "Opened!\n");
    return (0);			/* success */
}

static int
qikdata_release(struct inode *inode, struct file *filp)
{
    PHW_DEVICE_EXTENSION devExt =
	&qikdata_device_extension[QD_HBA_NR(inode->i_rdev)];
    unsigned long iflags;
    int sync_now;

    iflags = 0;
    HDEBUG(KERN_ALERT HERE "spin_lock_irqsave(&devExt->lock);\n");
    spin_lock_irqsave(&devExt->lock, iflags);
    devExt->usage--;
    sync_now = !devExt->usage;
    spin_unlock_irqrestore(&devExt->lock, iflags);
    if (sync_now)
    {
	/*
	 * but flush it right now
	 *
         * Note: this can cause qikdata_request to be invoked, to write out
         * the buffers, so be careful with spinlocks.
	 */
	fsync_dev(inode->i_rdev);
	invalidate_buffers(inode->i_rdev);
    }
    MOD_DEC_USE_COUNT;
    HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&devExt->lock);\n");
    return (0);
}

/* The ioctl() implementation */

static int
qikdata_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
    unsigned long arg)
{
    int     err;
    int     cyls;
    int     sects;
    u_long  result;
    int     hba_idx = QD_HBA_NR(inode->i_rdev);
    int     drv_idx = QD_DRV_NR(inode->i_rdev);

    HDEBUG(KERN_ALERT HERE "ioctl 0x%X 0x%lX\n", cmd, arg);
    if (hba_idx >= qd_devs)
	return -ENODEV;
    switch (cmd)
    {
    case BLKGETSIZE:
	HDEBUG(KERN_ALERT HERE "BLKGETSIZE\n");
	/*
	 * Return the device size, expressed in sectors
	 */
	if (!arg)
	    return -EINVAL;	/* NULL pointer: not valid */
	err = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long));
	if (err)
	    return err;
	result = qikdata_gendisk.part[MINOR(inode->i_rdev)].nr_sects;
	HDEBUG(KERN_ALERT HERE "result = %ld\n", result);
	put_user(result, (long *)arg);
	return 0;

    case BLKFLSBUF:
	/*
	 * flush
	 */
	HDEBUG(KERN_ALERT HERE "BLKFLSBUF\n");
	if (!suser())
	    return -EACCES;	/* only root */
	fsync_dev(inode->i_rdev);
	invalidate_buffers(inode->i_rdev);
	return 0;

    case BLKRAGET:
	/*
	 * return the readahead value
	 */
	HDEBUG(KERN_ALERT HERE "BLKRAGET\n");
	if (!arg)
	    return -EINVAL;
	err = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long));
	if (err)
	    return err;
	put_user(read_ahead[MAJOR(inode->i_rdev)], (long *)arg);
	return 0;

    case BLKRASET:
	/*
	 * set the readahead value
	 */
	HDEBUG(KERN_ALERT HERE "BLKRASET\n");
	if (!suser())
	    return -EACCES;
	if (arg > 0xff)
	    return -EINVAL;	/* limit it */
	read_ahead[MAJOR(inode->i_rdev)] = arg;
	return 0;

    case BLKRRPART:
	/*
	 * re-read partition table: fake a disk change
	 */
	HDEBUG(KERN_ALERT HERE "BLKRRPART(%d:%d)\n", MAJOR(inode->i_rdev),
	    MINOR(inode->i_rdev));

	qikdata_dump_partition_table(inode->i_rdev, dump_buff);
	return (qikdata_revalidate(inode->i_rdev));

    case HDIO_GETGEO:
	{
	    struct hd_geometry *geo_p;
	    struct hd_geometry geo;

	    HDEBUG(KERN_ALERT HERE "HDIO_GETGEO\n");
	    geo_p = (struct hd_geometry *)arg;

	    /*
	     * Verify the return address.
	     */
	    if (geo_p == NULL)
		return -EINVAL;
	    err = verify_area(VERIFY_WRITE, geo_p, sizeof(*geo_p));
	    if (err)
		return err;

	    /*
	     * Fake the geometry the same way as the aha152x driver
	     * ie. hd = FAKE_HEADS, s = FAKE_SECTORS,
	     * cyl = cap / (FAKE_HEADS*FAKE_SECTORS)
	     *
	     * assume 512-byte hard sectors for the moment...
	     * qd_sizes is in 1KB blocks, so double for total sectors.
	     *
	     * Round the minor device number down to zero, so that we get
	     * the size of the whole disk, not just the current partition.
	     */
	    sects = qd_sizes[QD_MK_MIN(hba_idx, drv_idx, 0)] * 2;
	    cyls = sects / (FAKE_HEADS * FAKE_SECTORS);

	    /*
	     * Build the result,
	     * and copy it into user space.
	     */
	    memset(&geo, 0, sizeof(geo));
	    if (cyls >= (1uL << 16))
		geo.cylinders = (1uL << 16) - 1;
	    else
		geo.cylinders = cyls;
	    geo.heads = FAKE_HEADS;
	    geo.sectors = FAKE_SECTORS;
	    geo.start = 0;
	    HDEBUG(KERN_ALERT HERE "Cyl=%d, Hed=%d, Sec=%d, Sta=%ld\n",
		geo.cylinders, geo.heads, geo.sectors, geo.start);
	    copy_to_user(geo_p, &geo, sizeof(geo));
	}
	return (0);

#ifdef HDIO_GETGEO_BIG
    case HDIO_GETGEO_BIG:
	{
	    struct hd_big_geometry *geo_p;
	    struct hd_big_geometry geo;

	    HDEBUG(KERN_ALERT HERE "HDIO_GETGEO_BIG\n");
	    geo_p = (struct hd_big_geometry *)arg;

	    /*
	     * Verify the return address.
	     */
	    if (geo_p == NULL)
		return -EINVAL;
	    err = verify_area(VERIFY_WRITE, geo_p, sizeof(*geo_p));
	    if (err)
		return err;

	    /*
	     * Fake the geometry the same way as the aha152x driver
	     * ie. hd = FAKE_HEADS, s = FAKE_SECTORS,
	     * cyl = cap / (FAKE_HEADS*FAKE_SECTORS)
	     *
	     * assume 512-byte hard sectors for the moment...
	     * qd_sizes in in 1KB blocks, so double for total sectors.
	     *
	     * Round the minor device number down to zero, so that we get
	     * the size of the whole disk, not just the current partition.
	     */
	    sects = qd_sizes[QD_MK_MIN(hba_idx, drv_idx, 0)] * 2;
	    cyls = sects / (FAKE_HEADS * FAKE_SECTORS);

	    /*
	     * Build the result,
	     * and copy it into user space.
	     */
	    memset(&geo, 0, sizeof(geo));
	    geo.cylinders = cyls;
	    geo.heads = FAKE_HEADS;
	    geo.sectors = FAKE_SECTORS;
	    geo.start = 0;
	    HDEBUG(KERN_ALERT HERE "Cyl=%d, Hed=%d, Sec=%d, Sta=%ld\n",
		geo.cylinders, geo.heads, geo.sectors, geo.start);
	    copy_to_user(geo_p, &geo, sizeof(geo));
	}
	return (0);
#endif

    case QIKDRIVE_IOCTL_CMD:
	return qikdata_handle_ioctl(inode->i_rdev, cmd, arg);

    case QD_IOCTL_HELLO:
	{
	    qd_ioctl_hello_t result;
	    qd_ioctl_hello_t *user;

	    /*
	     * Verify the return address.
	     */
	    HDEBUG(KERN_ALERT HERE "QD_IOCTL_HELLO\n");
	    user = (qd_ioctl_hello_t *) arg;
	    if (user == NULL)
		return -EINVAL;
	    err = verify_area(VERIFY_WRITE, user, sizeof(*user));
	    if (err)
		return err;

	    /*
	     * Build the result.
	     */
	    strncpy(result.signature, QD_IOCTL_HELLO_SIGNATURE,
		sizeof(result.signature));
	    result.instance = inode->i_rdev >> QD_SHIFT;

	    /*
	     * Copy the result back into user space.
	     */
	    copy_to_user(user, &result, sizeof(result));
	}
	return 0;

    default:
	HDEBUG(KERN_ALERT HERE "generic block device ioctl\n");
	return (blk_ioctl(inode->i_rdev, cmd, arg));
    }

    return -EINVAL;		/* unknown command */
}

/*  Support for removable devices */

static int
qikdata_check_change(kdev_t i_rdev)
{
    int     minor = MINOR(i_rdev);

    if (minor >= qd_devs)	/* paranoid */
	return 0;

    /*
     * return 1 here if we need to change media
     */
    return 0;			/* still valid */
}

/* The file operations */

#ifdef LINUX_22
#define block_device_operations file_operations
#endif

static struct block_device_operations qikdata_fops =
{
    open: qikdata_open,
    release: qikdata_release,
    ioctl: qikdata_ioctl,
    revalidate: qikdata_revalidate,
    check_media_change: qikdata_check_change,
#ifndef LINUX_24
    read: block_read,
    write: block_write,
    fsync: block_fsync,
#endif
};

/*
 * here we need to re-read the partition table and call resetup_one_dev
 */
static int
qikdata_revalidate(kdev_t i_rdev)
{
    int     dev = DEVICE_NR(i_rdev);
    int     majnr = MAJOR(i_rdev);
    int     minnr0 = MINOR_NR(dev, 0);
    struct hd_struct *hd = &qikdata_gendisk.part[minnr0];
    int     part;

    HDEBUG(KERN_ALERT HERE "revalidate devnr[%d:%d]\n", majnr, MINOR(i_rdev));

    /*
     * fill in the blksizes, hardsects arrays manually
     */
    for (part = 1; part < (1 << QD_SHIFT); part++)
    {
	int     prt_minr;

	prt_minr = MINOR_NR(dev, part);
	if (hd[part].nr_sects != 0)
	{
	    kdev_t  devp;
	    struct super_block *sb;

	    devp = MKDEV(majnr, prt_minr);
	    sb = get_super(devp);
	    fsync_dev(devp);
	    if (sb)
		invalidate_inodes(sb);
	    invalidate_buffers(devp);
	    hd[part].nr_sects = 0;
	}
	/*
	 * zero the old information
	 */
	hd[part].start_sect = 0;
	qikdata_gendisk.sizes[prt_minr] = 0;
	/*
	 * copy values from the whole device to the partition
	 */
	qd_blocksizes[prt_minr] = qd_blocksizes[minnr0];
	qd_hardsects[prt_minr] = qd_hardsects[minnr0];
    }
    /*
     * then fill in new info
     */
    HDEBUG(KERN_ALERT HERE "partition check:\n");
#ifdef LINUX_24
    HDEBUG(KERN_ALERT HERE "qd_partitions[minnr0].nr_sects = %ld\n",
	qd_partitions[minnr0].nr_sects);

    qikdata_dump_gendisk(&qikdata_gendisk, i_rdev, QD_MAXRDEV,
	(void *)&qikdata_fops, qd_partitions[minnr0].nr_sects);

    HDEBUG(KERN_ALERT HERE "register_disk {\n");
    register_disk(&qikdata_gendisk, i_rdev, QD_MAXRDEV, (void *)&qikdata_fops,
	qd_partitions[minnr0].nr_sects);
    HDEBUG(KERN_ALERT HERE "register_disk }\n");
#else
    HDEBUG(KERN_ALERT HERE "resetup_one_dev {\n");
    resetup_one_dev(&qikdata_gendisk, dev);
    HDEBUG(KERN_ALERT HERE "resetup_one_dev }\n");
#endif
    HDEBUG(KERN_ALERT HERE "partition check done\n");

    return 0;
}

/*
 * qd_end_request - provides a wrapper for end_request to deal with
 * the situation where CURRENT is not correct.  This happens during
 * driver initialisation or possibly a kernel bug.   A better approach
 * is needed for the initialisation case, probably in the isr.
 */
static void
qd_end_request(int uptodate, PHW_DEVICE_EXTENSION devExt)
{
    struct request *req = CURRENT;

    if (devExt->major != MAJOR(req->rq_dev))
    {
	HDEBUG(KERN_ALERT HERE "devExt->major != MAJOR(req->rq_dev)\n");
	return;
    }
    HDEBUG(KERN_ALERT HERE "about to end_request(%d) CURRENT=%p\n", uptodate,
	CURRENT);
    end_request(uptodate);
}


static void
#ifdef LINUX_24
qikdata_request(request_queue_t *rqueue)
#else
qikdata_request(void)
#endif
{
    PHW_DEVICE_EXTENSION devExt;
    int     size;
    int     minor;
    int     hba_idx;
    kdev_t  i_rdev;
    unsigned long flags;

    while (1)
    {
	/*
	 * The INIT_REQUEST macro will return (!) from this function if
	 * the queue is empty.	It will panic if the queue has become
	 * corrupted.
	 */
	INIT_REQUEST;

	HDEBUG(KERN_ALERT HERE "qikdata_request: CURRENT=%p\n", CURRENT);

	i_rdev = CURRENT->rq_dev;
	hba_idx = QD_HBA_NR(i_rdev);
	minor = MINOR(i_rdev);
	devExt = &(qikdata_device_extension[hba_idx]);
	HDEBUG(KERN_ALERT HERE "spin_lock_irqsave(&devExt->lock);\n");
	spin_lock_irqsave(&devExt->lock, flags);

	/*
	 * Check the device number is in range
	 */
	if (hba_idx >= qd_devs)
	{
	    static int count = 0;

	    if (count++ < 5)	/* print the message at most five times */
		HDEBUG(KERN_WARNING HERE "request for unknown device\n");
	    qd_end_request(0, devExt);
	    goto another_one;
	}

	/*
	 * calculate the size in sectors so we don't get overflow
	 * - note that this code assumes 512-byte sectors
	 */
	size = CURRENT->current_nr_sectors;

	if ((CURRENT->sector + size) > qikdata_gendisk.part[minor].nr_sects)
	{
	    static int count = 0;

	    if (count++ < 5)
		HDEBUG(KERN_WARNING HERE "request past end of device\n");
	    qd_end_request(0, devExt);
	    goto another_one;
	}

	/*
	 * catch the largest transfer size
	 */
	if (size > devExt->LargestTransferSize)
	    devExt->LargestTransferSize = size;

	/*
	 * check for alignment
	 */
	if (CURRENT->sector & (CURRENT->current_nr_sectors - 1))
	{
	    HDEBUG(KERN_ALERT HERE "error: received unaligned transfer"
		" (sector=%lX, size=%lX)\n", CURRENT->sector,
		CURRENT->current_nr_sectors);
	    devExt->ReceivedUnaligned = 1;
	}
	devExt->State = DEV_STATE_DATA;

	switch (CURRENT->cmd)
	{
	case READ:
	    devExt->PendingRequest = 1;
	    qikdata_handle_read(i_rdev, CURRENT->buffer,
		qd_partitions[minor].start_sect + CURRENT->sector,
		CURRENT->current_nr_sectors);
    	    HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&devExt->lock);\n");
	    spin_unlock_irqrestore(&devExt->lock, flags);
	    return;

	case WRITE:
	    HDEBUG(KERN_ALERT HERE "mark\n");
	    devExt->PendingRequest = 1;
	    HDEBUG(KERN_ALERT HERE "about to qikdata_handle_write\n");
	    qikdata_handle_write(i_rdev, CURRENT->buffer,
		qd_partitions[minor].start_sect + CURRENT->sector,
		CURRENT->current_nr_sectors);
	    HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&devExt->lock);\n");
	    spin_unlock_irqrestore(&devExt->lock, flags);
	    HDEBUG(KERN_ALERT HERE "mark\n");
	    return;

	default:
	    /*
	     * can't happen
	     */
	    devExt->PendingRequest = 0;
	    printk(KERN_ALERT HERE "error: unhandled CURRENT->cmd\n");
	    if (CURRENT == NULL)
		printk(KERN_ALERT HERE "CURRENT is NULL\n");
	    else
		qd_end_request(0, devExt);
	    break;
	}
    another_one:
	HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&devExt->lock);\n");
	spin_unlock_irqrestore(&devExt->lock, flags);
    }
}


static int
qikdata_find_hba(int devIndex, u16 QikDataType)
{
    PHW_DEVICE_EXTENSION devExt = NULL;
    struct pci_dev *pdev = NULL;
    int     pci_index = 0;
    int     err = 0;

    /*
     * No spinlocks required
     */
    HDEBUG(KERN_ALERT HERE "qikdata_find_hba:\n");

    if (!pci_present())
	return devIndex;
    for (pci_index = 0; pci_index < QD_MAXRDEV; pci_index++)
    {
	unsigned int pci_reg;
	int     i;

	HDEBUG(KERN_ALERT HERE "qikdata_find_hba: devIndex = %d, "
	    "pci_index = %d\n", devIndex, pci_index);

	/*
	 * Only allow QD_MAXRDEV devices
	 */
	if ((devIndex) >= QD_MAXRDEV)
	{
	    HDEBUG(KERN_ALERT HERE "Found %d devices.  QD_MAXRDEV = %d\n",
		devIndex + pci_index, QD_MAXRDEV);
	    err = 1;
	    break;
	}

	/*
	 * get pointer to device extension
	 */
	devExt = &qikdata_device_extension[devIndex];
	pdev = pci_find_device(VENDOR_PLATYPUS, QIKDATA_HBA_CARD, pdev);
	if (!pdev)
	{
	    HDEBUG(KERN_ALERT HERE "pci_find_device - no more devices. "
		"found = %d\n", pci_index);
	    err = 2;
	    break;
	}
	devExt->pci_dev = pdev;
	HDEBUG(KERN_ALERT HERE "Found a QikDATA HBA card\n");
#ifdef LINUX_24
	err = pci_enable_device(devExt->pci_dev);
	if (err)
	{
	    HDEBUG(KERN_ALERT HERE "pci_enable_device failed:%d\n", err);
	    err = err + 2;
	    break;
	}
#else
	err = 0;
#endif

	/*
	 * Set the array index in the device extension
	 */
	devExt->DeviceIndex = devIndex;

	/*
	 * Determine the qd type
	 */
	devExt->QikDriveType = devExt->pci_dev->device;

	/*
	 * Extract the assigned interrupt
	 */
	devExt->irq = devExt->pci_dev->irq;
	HDEBUG(KERN_ALERT HERE "IRQ = %d\n", (int)devExt->irq);
	devExt->PendingRequest = PENDING_REQUEST_BOGUS;

	pci_reg = PCI_BASE_ADDRESS_0;
	for (i = 0; i < 2; i++)
	{
#ifdef LINUX_24
	    devExt->BAR[i] = pci_resource_start(pdev, i);
#else
	    devExt->BAR[i] = pdev->base_address[i];
#endif

	    /*
	     * determine the range of address window
	     */
	    cli();
	    pci_write_config_dword(devExt->pci_dev, pci_reg, ~0);
	    pci_read_config_dword(devExt->pci_dev, pci_reg,
		&(devExt->BARrange[i]));
	    pci_write_config_dword(devExt->pci_dev, pci_reg, devExt->BAR[i]);
	    sti();

	    if ((devExt->BAR[i] & PCI_BASE_ADDRESS_SPACE) ==
		PCI_BASE_ADDRESS_SPACE_IO)
	    {
		devExt->BARtype[i] = PCI_BASE_ADDRESS_SPACE_IO;
		devExt->BAR[i] &= PCI_BASE_ADDRESS_IO_MASK;
		HDEBUG(KERN_ALERT HERE "BAR%d = %08lX [io] ", i,
		    devExt->BAR[i]);
	    }
	    else
	    {
		devExt->BARtype[i] = PCI_BASE_ADDRESS_SPACE_MEMORY;
		devExt->BAR[i] &= PCI_BASE_ADDRESS_MEM_MASK;
		HDEBUG(KERN_ALERT HERE "BAR%d = %08lX [mem]", i,
		    devExt->BAR[i]);
	    }
	    devExt->BARrange[i] = ~(devExt->BARrange[i] & ~1) + 1;
	    HDEBUG("  %8d bytes\n", devExt->BARrange[i]);

	    pci_reg += 4;
	}

	/*
	 * Grab a virtual pointer to the local address space memory window
	 */
	devExt->LASmap = ioremap(devExt->BAR[LAS], devExt->BARrange[LAS]);
	HDEBUG(KERN_ALERT HERE "Virtual Address for LAS (SDRAM) = $%08X\n",
	    (u_int) (devExt->LASmap));

	/*
	 * Grab a virtual pointer to the register window
	 */
	devExt->RTRmap = ioremap(devExt->BAR[RTR], devExt->BARrange[RTR]);
	HDEBUG(KERN_ALERT HERE "Virtual Address for RTR (registers) = $%08X\n",
	    (u_int) (devExt->RTRmap));

	/*
	 * Increment the number of devices
	 */
	devIndex++;
    }
    if (err > 2)
	devIndex = 0;
    return (devIndex);
}

/**
 * check_compatibility
 * INPUT: PHW_DEVICE_EXTENSION
 *        minimum_compatibility
 *
 * OUPUT: int
 *
 * Function checks that each drive on the ring is compatible with the
 * highest earliest version encountered in the CABLE_E_EARLIEST_COMPATIBLE
 * cable register, then it checks that the QikDATA HBA version is
 * compatible with the QikDATA drive versions.
 */
static int
check_compatibility(PHW_DEVICE_EXTENSION devExt, int min_compatibility)
{
    int     status = 1;
    int     n;

    if (devExt->nBays > 0)
    {
	for (n = 0; n < MAX_DRIVES; n++)
	{
	    if ((devExt->UnitDet[n].drive_type != 0) &&
		(devExt->UnitDet[n].UnitReady) &&
		(devExt->UnitDet[n].pack_version < min_compatibility))
	    {
		HDEBUG(KERN_ALERT HERE "QikData[%d] has pack_version %d\n", n,
		    devExt->UnitDet[n].pack_version);
		status = 0;
		HDEBUG(KERN_ALERT HERE "Incompatible packs!\n");
	    }
	}
	if (((devExt->AcexIDReg & 0xFFFF) >= 0xAD2) && (min_compatibility < 19))
	{
	    status = 0;
	    HDEBUG(KERN_ALERT HERE "QikDATA HBA and QikDATA Drive are "
		"incompatible\n");
	    HDEBUG(KERN_ALERT HERE "QikDATA HBA version 0x%X, minimum Drive "
		"version %d\n",
		devExt->AcexIDReg, min_compatibility);
	}
	if (((devExt->AcexIDReg & 0xFFFF) < 0xAD2) && (min_compatibility >= 19))
	{
	    HDEBUG(KERN_ALERT HERE "QikDATA HBA and QikDATA Drive are "
		"incompatible\n");
	    HDEBUG(KERN_ALERT HERE "QikDATA HBA version 0x%X, minimum Drive "
		"version %d\n",
		devExt->AcexIDReg, min_compatibility);
	    status = 0;
	}

    }
    HDEBUG("check compatibility: min_comp %d, returning %d.\n",
	min_compatibility, status);
    return (status);
}


static void
HwReset(int device_number)
{
    PHW_DEVICE_EXTENSION devExt;
    unsigned long dma;
    unsigned long flags;
    unsigned pRTR;

    devExt = &(qikdata_device_extension[device_number]);
    HDEBUG(KERN_ALERT HERE "spin_lock_irqsave(&devExt->lock);\n");
    spin_lock_irqsave(&devExt->lock, flags);

    pRTR = (unsigned)devExt->RTRmap;

    /*
     * Clear the page register.
     */
    writel(0, pRTR + REG_PAGE);

    /*
     * tell if any interrupts still pending
     */
    dma = readl(pRTR + REG_DMA);
    if (dma & DMA_INTERRUPT_PENDING)
	printk(KERN_ALERT HERE "Interrupt still pending at cleanup\n");

    if (dma & DMA_PCI_CORE_INTERRUPT)
    {
	unsigned pciISR;

	printk(KERN_ALERT HERE "PCIAbortInterruptActive\n");

	/*
	 * Clear all possible PCI-related interrupts
	 */
	pciISR =
	    (PCI_STATUS_DATA_PARITY_DETECTED |
	    PCI_STATUS_SIGNALED_TARGET_ABORT |
	    PCI_STATUS_RECEIVED_TARGET_ABORT |
	    PCI_STATUS_RECEIVED_MASTER_ABORT |
	    PCI_STATUS_SIGNALED_SYSTEM_ERROR |
	    PCI_STATUS_DETECTED_PARITY_ERROR);
	pci_write_config_word(devExt->pci_dev, REG_PCIISR, pciISR);
    }

    /*
     * empty the fifo
     * clear all possible interrupts
     */
    HDEBUG(ISR_DEBUG HERE "dma = %s (0x%08X)\n", dma_decode(dma), dma);
    writel(dma | DMA_FIFO_CLEAR | DMA_SGL_INTERRUPT | DMA_ABORT_INTERRUPT |
	DMA_PCI_CORE_INTERRUPT | DMA_ECC_INTERRUPT, pRTR + REG_DMA);
    HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&devExt->lock);\n");
    spin_unlock_irqrestore(&devExt->lock, flags);
}


static int
HwInitialise(kdev_t i_rdev)
{
    /*
     * FUTURE CHANGE:  this note is to record thoughts for possible future
     * changes.
     *
     * 1. Change parameter devExt to a kdev_t progressively over next few
     *    changes!
     * 2. Extract drv_idx with QD_DRV_NR if (drv_idx == 0) initialize all
     *    drives, else init one drive (ie. hot-plug or error recovery). Maybe
     *    require mod to all code referencing device no.
     *
     */
    int     hba_idx = QD_HBA_NR(i_rdev);
    int     drv_idx = QD_DRV_NR(i_rdev);
    PHW_DEVICE_EXTENSION devExt = &(qikdata_device_extension[hba_idx]);
    u_int   pRTR = (u_int) devExt->RTRmap;
    u_int   pMEM = (u_int) devExt->LASmap;
    PAGE_REG pageRegister;
    int     i;
    int     rc = 0;
    u_long  dma;
    u_long  read_val;

    /*
     * this records the lowest pack that each QikData is compatible with
     */
    int     minimum_compatibility = 19;

    /*
     * Initialise the ACEX I/O register
     */
    devExt->AcexIOReg = 0;

    /*
     * Read the ACEX identification register
     */
    devExt->AcexIDReg = readl(pRTR + REG_ID);
    HDEBUG(KERN_ALERT HERE "QikDATA HBA version = %X\n", devExt->AcexIDReg);

    /*
     * Reset the hardware.
     */
    HwReset(hba_idx);

    /*
     * Read the IO register to save if we are 32/64bit and 33/66mhz and the
     * local bus clock speed (not implemented yet, for diagnostic purposes)
     */
    read_val = readl(pRTR + REG_IO);

    devExt->DeviceReady = 0;

    /*
     * Identify the drive units attached.
     */
    devExt->nBays = 0;
    drv_idx = 0;
    for (i = 1; i < (MAX_QD_NR + 1); i++)
    {
	PUNIT_DETAILS unitDet;
	kdev_t  drv_rdev;

	devExt->UnitIndex = drv_idx;	/* for 2.2 compatibility */
	unitDet = &(devExt->UnitDet[drv_idx]);
	unitDet->drive_address = i;
	drv_rdev = MKDEV(devExt->major, QD_MK_MIN(hba_idx, drv_idx, 0));
	sprintf(unitDet->drive_name, "%s%c", devExt->qd_name, 'a' + drv_idx);
	unitDet->drive_type = (u_long) 0;

	/*
	 * send a link reset
	 */
	qikdata_link_reset(drv_rdev);

	/*
	 * Clear previous interrupts.
	 */
	pageRegister.dw =
	    (DMA_SGL_INTERRUPT | DMA_FIFO_CLEAR | DMA_ABORT_INTERRUPT |
	    DMA_INTERRUPT_PENDING);
	writel(pageRegister.dw, pRTR + REG_DMA);

	target_abort = 0;

	/*
	 * Do a read from nios space to see if there is a drive there
	 * NOTE: this target-mode method needs to be replaced with DMA-mode!
	 */
	pageRegister.dw = (PAGE_MASK_NIOS | PAGE_MASK_EXTENDED);
	pageRegister.s.drive = unitDet->drive_address;	/* 6-bit field */
	writel(pageRegister.dw, pRTR + REG_PAGE);
	unitDet->drive_type = readl(pMEM + CABLE_E_MAGIC);
	HDEBUG(KERN_ALERT HERE "unitDet->drive_type = 0x%08lX\n",
	    unitDet->drive_type);

	/*
	 * If we didn't timeout there is a drive there.
	 *
	 * BUT: on a sufficiently fast machine, this poll can come too early!
	 */
	udelay(17);
	dma = readl(pRTR + REG_DMA);

	/*
	 * Check that if an interrupt occured it was not a dma abort.
	 */
	if (!(dma & DMA_INTERRUPT_PENDING))
	{
	    HDEBUG(KERN_ALERT HERE "interrupt pending.\n");
	    if (target_abort)
	    {
		HDEBUG(KERN_ALERT HERE "DMA ABORT\n");
		unitDet->drive_type = 0;
	    }
	}
	else
	{
	    unitDet->drive_type = 0;
	}

	if (unitDet->drive_type == (u_long)DRIVE_MAGIC)
	{
	    /*
	     * Read DIMM geometry
	     */
	    qikdata_read_dimm_spds(drv_rdev);

	    /*
	     * Initialise dimm statistics
	     */
	    qikdata_initialize_dimm_statictics(drv_rdev);

	    /*
	     * If everything is OK set the drive unit ready
	     */
	    unitDet->UnitReady = 1;

	    /*
	     * Set drive unit. Tests if multi-drive ok
	     */
	    devExt->nBays++;
	    HDEBUG(KERN_ALERT HERE "hba[%d].drive[%d].addr = %d,"
		" type = 0x%08X, name = %s\n", hba_idx, drv_idx,
		unitDet->drive_address, (int)unitDet->drive_type,
		unitDet->drive_name);
	    drv_idx++;

#ifdef QIKDATA_DEBUG
	    /*---- DEBUG TRACE -  DELETE WHEN NOT RELEVANT ---------------*/
	    {
		int j;
		for (j = MAX_DIMMS - 1; j >= 0; j--)
		{
		    if (unitDet->Dimm[j].assembly_serial == -1)
			continue;

		    printk(KERN_ALERT HERE
			"Dimm XF%d:%5dMB %-3.3s %-8.8s Serial#:%010u\n", j,
			(int)(qikdata_dimm_capacity(&(unitDet->Dimm[j])) >> 20),
			(unitDet->Dimm[j].ecc == 0x02 ? "ECC" : ""),
			unitDet->Dimm[j].manufacturer_jedec_id,
			unitDet->Dimm[j].assembly_serial);
		}
	    }
	    /*-----------------------------------------------------------*/
#endif /* QIKDATA_DEBUG */

	    /*
	     * Read & store the CABLE_E_PACK_VERSION cable register and
	     * the CABLE_E_EARLIEST_COMPATIBLE_PACK cable register
	     * from each QikData.
	     */
	    pageRegister.dw = (PAGE_MASK_NIOS | PAGE_MASK_EXTENDED);
	    pageRegister.s.drive = unitDet->drive_address; /* 6-bit field */
	    writel(pageRegister.dw, pRTR + REG_PAGE);

	    unitDet->pack_version = readl(pMEM + CABLE_E_PACK_VERSION);
	    unitDet->earliest_compatible_pack =
		readl(pMEM + CABLE_E_EARLIEST_COMPATIBLE_PACK);
	    HDEBUG(KERN_ALERT HERE "For QikData %d pack version is %d.\n",
		unitDet->drive_address, unitDet->pack_version);
	    HDEBUG(KERN_ALERT HERE "For QikData %d earliest compatible pack\n",
		unitDet->drive_address);
	    HDEBUG(KERN_ALERT HERE "version is %d.\n",
		unitDet->earliest_compatible_pack);

	    /*
	     * remember the smallest pack that the QikData is compatible with
	     */
	    if (unitDet->earliest_compatible_pack < minimum_compatibility)
		minimum_compatibility = unitDet->earliest_compatible_pack;

	    /*
	     * Signal to the drive that we are ready
	     */
	    HDEBUG(KERN_ALERT HERE "signal drive %d we are loading\n",
		unitDet->drive_address);
	    writel(1, pMEM + CABLE_WE_DRVLOAD);
	    writel(1, pMEM + CABLE_WE_DRVLOAD + 4);
	}
	else
	{
	    unitDet->UnitReady = 0;
	    HDEBUG(KERN_ALERT HERE "drive[%d] not attached!\n",
		unitDet->drive_address);
	    unitDet->drive_address = 0;
	}

	/*
	 * reset the page register
	 */
	writel(0, pRTR + REG_PAGE);
    }
    if (devExt->nBays > 0)
	devExt->DeviceReady = 1;	/* at least 1 drive. */

    /*
     * clear system ??
     */
    qikdata_link_reset(i_rdev);
    pageRegister.dw =
	(DMA_SGL_INTERRUPT | DMA_FIFO_CLEAR | DMA_ABORT_INTERRUPT |
	DMA_INTERRUPT_PENDING);
    writel(pageRegister.dw, pRTR + REG_DMA);
    writel(0, pRTR + REG_PAGE);

    if (devExt->DeviceReady)
    {
	/*
	 * Now turn off the attention LED
	 */
	devExt->AcexIOReg &= ~IO_nLED_ATTN;

	/*
	 * if qikdata_cable_ecc is set, enable cable ecc
	 */
	if (qikdata_cable_ecc)
	{
	    devExt->AcexIOReg |= IO_DAT_ECC | IO_CMD_ECC;
	    HDEBUG(KERN_ALERT HERE "Enabling ECC on QikDATA HBA %d\n"
		devExt->DeviceIndex);
	}
	writel(devExt->AcexIOReg, pRTR + REG_IO);
	rc = 0;
    }
    else
    {
	HDEBUG(KERN_ALERT HERE "No units found on QikDATA HBA %d\n",
	    devExt->DeviceIndex);

	rc = 1;
    }

    /*
     * Check if all of the QikDATA Drives and the QikDATA HBAs in the ring are
     * compatible. If they are not then fail attach.
     */
    if (!check_compatibility(devExt, minimum_compatibility))
    {
	printk(KERN_ALERT "QikDATA Drive and QikDATA HBA versions not"
	    " compatible!\n");
	rc = 1;
    }
    return (rc);
}

static void
HwInterrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    kdev_t  devnr = (kdev_t) ((ulong) dev_id);

    PHW_DEVICE_EXTENSION devExt;
    u_int   pRTR;
    u32     dma;
    kdev_t  i_rdev = 0;
    int     hba_idx;
    int     drv_idx;
    int     majnr = MAJOR(devnr);
    int     state;
    int     ok = 0;
    int     iflags = 0;
    int     dma_abort_intr = 0;

    /*
     * We take a lock around all uses of the CURRENT pointer,
     * and all exit paths from this function release it again.
     */
    HDEBUG(KERN_ALERT HERE "spin_lock_irqsave(&io_request_lock);\n");
    spin_lock_irqsave(&io_request_lock, iflags);

    if (CURRENT)
    {
	i_rdev = CURRENT->rq_dev;
	hba_idx = QD_HBA_NR(i_rdev);
	drv_idx = QD_DRV_NR(i_rdev);
	devExt = &(qikdata_device_extension[hba_idx]);
	state = 1;
    }
    else
    {
	hba_idx = QD_HBA_NR(devnr);
	devExt = &(qikdata_device_extension[hba_idx]);
	drv_idx = devExt->UnitIndex;	/* used during initialisation */
	i_rdev = MKDEV(majnr, QD_MK_MIN(hba_idx, drv_idx, 0));
	state = 0;
    }
    HDEBUG(KERN_ALERT HERE "Interrupt(): devnr[%d:%d], i_rdev[%d:%d], "
	"state = %d\n", MAJOR(devnr), MINOR(devnr), MAJOR(i_rdev),
	MINOR(i_rdev), state);

    /*
     * Do the sources of major number and hba_idx agree
     */
    if ((devExt->PendingRequest != PENDING_REQUEST_BOGUS) &&
	((devExt->major != majnr) || (MAJOR(i_rdev) != majnr)))
    {
	printk(KERN_ALERT HERE "wrong major: devnr[%d:%d], i_rdev[%d:%d],"
	    " devExt->major = %d\n", MAJOR(devnr), MINOR(devnr),
	    MAJOR(i_rdev), MINOR(i_rdev), devExt->major);
	printk(KERN_ALERT " && not bogus request, state = %d\n\n", state);
	goto done;
    }
    if (hba_idx != QD_HBA_NR(i_rdev) && CURRENT)
    {
	printk(KERN_ALERT HERE "wrong QikDATA HBA (%d != %d)\n",
	    hba_idx, QD_HBA_NR(i_rdev));
	goto done;
    }
    HDEBUG(KERN_ALERT HERE "spin_lock(&devExt->lock);\n");
    spin_lock(&devExt->lock);
    pRTR = (u_int) devExt->RTRmap;
    dma = readl(pRTR + REG_DMA);
    HDEBUG(ISR_DEBUG HERE "dma = %s (0x%08X)\n", dma_decode(dma), dma);

    /*
     * return immediately if we didn't cause the interrupt
     */
    if (!(dma & DMA_INTERRUPT_PENDING))
    {
	HDEBUG(KERN_ALERT HERE "spin_unlock(&devExt->lock);\n");
	spin_unlock(&devExt->lock);
	goto done;
    }

#ifdef FLASHER
    /*
     * turn ATTN LED on
     */
    devExt->AcexIOReg |= IO_nLED_ATTN;
    writel(devExt->AcexIOReg, pRTR + REG_IO);
#endif

    if (dma & DMA_ABORT_INTERRUPT)
    {
	dma_abort_intr = 1;
	HDEBUG(KERN_ALERT HERE "dma abort interrupt.\n");
    }

    /*
     * Clear any pending PCI ABORT sources
     */
    if (dma & DMA_PCI_CORE_INTERRUPT)
    {
	ushort  pciISR;

	printk(KERN_ALERT HERE "PCIAbortInterruptActive\n");
	/*
	 * Clear all possible PCI-related interrupts
	 */
	pciISR =
	    (PCI_STATUS_DATA_PARITY_DETECTED |
	    PCI_STATUS_SIGNALED_TARGET_ABORT |
	    PCI_STATUS_RECEIVED_TARGET_ABORT |
	    PCI_STATUS_RECEIVED_MASTER_ABORT |
	    PCI_STATUS_SIGNALED_SYSTEM_ERROR |
	    PCI_STATUS_DETECTED_PARITY_ERROR);
	pci_write_config_word(devExt->pci_dev, REG_PCIISR, pciISR);
    }

    /*
     * clear all QikDATA HBA interrupts
     */
    writel(dma | DMA_SGL_INTERRUPT | DMA_ABORT_INTERRUPT |
	DMA_PCI_CORE_INTERRUPT | DMA_ECC_INTERRUPT, pRTR + REG_DMA);

    /*
     * did we get an abort on the transaction?
     */
    if (dma_abort_intr)
    {
	if (devExt->PendingRequest == PENDING_REQUEST_BOGUS)
	{
	    state = 0;
	}
	else
	{
	    if (CURRENT)
	    {
		state = 1;
	    }
	    else
	    {
		state = -1;
		HDEBUG(KERN_ALERT HERE "CURRENT = NULL , but not initial.\n");
	    }
	}
	HDEBUG(KERN_ALERT HERE "Abort: attaching hba[%d].drive[%d].addr = %d,"
	    " state = %d, devnr[%d:%d]\n", hba_idx, QD_DRV_NR(i_rdev),
	    devExt->UnitDet[QD_DRV_NR(i_rdev)].drive_address, state,
	    MAJOR(i_rdev), MINOR(i_rdev));

	/*
	 * abort interrupt may signal a target abort!
	 */
	target_abort = 1;

	/*
	 * send a link reset
	 */
	qikdata_link_reset(i_rdev);
	ok = 0;
    }

    HDEBUG(KERN_ALERT HERE "spin_unlock(&devExt->lock);\n");
    spin_unlock(&devExt->lock);

    /*
     * Make sure we are waiting for a request to finish!
     */
    if (devExt->PendingRequest
	&& (devExt->PendingRequest != PENDING_REQUEST_BOGUS))
    {
	HDEBUG(KERN_ALERT HERE "looks like a valid request\n");

	/*
	 * for now we will assume everything worked...
	 */
	ok = 1;

	/*
	 * handle completion of pending request
	 */
	devExt->PendingRequest = 0;
	if (CURRENT)
	{
	    if (!ok)
	    {
		HDEBUG(KERN_WARNING HERE "end_request(0)");
	    }
	    iflags = 0;

	    qd_end_request(ok, devExt);

	    /*
	     * schedule any more pending requests
	     */
	    if (CURRENT)
	    {
#ifdef LINUX_24
		qikdata_request(BLK_DEFAULT_QUEUE(devExt->major));
#else
		qikdata_request();
#endif
	    }
	}
    }

#ifdef FLASHER
    /*
     * turn off ATTN LED
     */
    HDEBUG(KERN_ALERT HERE "spin_lock(&devExt->lock);\n");
    spin_lock(&devExt->lock);
    devExt->AcexIOReg &= ~IO_nLED_ATTN;
    writel(devExt->AcexIOReg, pRTR + REG_IO);
    HDEBUG(KERN_ALERT HERE "spin_unlock(&devExt->lock);\n");
    spin_unlock(&devExt->lock);
#endif

    done:
    HDEBUG(KERN_ALERT HERE "done handling this interrupt\n");
    HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&io_request_lock);\n");
    spin_unlock_irqrestore(&io_request_lock, iflags);
}

/* Finally, the module stuff */

int
qikdata_dump_partition_table(kdev_t i_rdev, char *buf)
{
    int     minnr;
    int     i;
    int     j;
    int     len = 0;
    int     hba_idx = QD_HBA_NR(i_rdev);
    PHW_DEVICE_EXTENSION devExt = &(qikdata_device_extension[hba_idx]);

#ifdef QIKDATA_DEBUG
    /*
     * display driver information
     */
    len += sprintf(buf + len, "major = %d\n", qikdata_gendisk.major);
    len += sprintf(buf + len, "major_name = \"%s\"\n",
	qikdata_gendisk.major_name);
    len += sprintf(buf + len, "minor_shift = %d\n",
	qikdata_gendisk.minor_shift);
    len += sprintf(buf + len, "max_p = %d\n", qikdata_gendisk.max_p);
    len += sprintf(buf + len, "nr_real = %d\n", qikdata_gendisk.nr_real);
#endif

    len += sprintf(buf + len, "\n");

    /*
     * display whole-disk information
     */
    for (i = 0; i < devExt->nBays; i++)
    {
	len += sprintf(buf + len,
	    "\n/dev/" DRIVER_LONG_NAME "/%s   hba[%d].drive[%d].address = %d\n",
	    devExt->UnitDet[i].drive_name, hba_idx, i,
	    devExt->UnitDet[i].drive_address);
	len += sprintf(buf + len,
	    "      %10.10s %10.10s %10.10s %10.10s %10.10s "
	    "%10.10s\n", "sizes", "start_sect", "nr_sects", "sizes",
	    "blk_sizes", "hardsects");
	/*
	 * show partitions for all drives attached to this HBA, in future
	 * revision only show all if drv_idx == 0 else show only one drive's.
	 */
	for (j = 0; j < (1 << QD_SHIFT); j++)
	{
	    minnr = QD_MK_MIN(hba_idx, i, j);
	    /*
	     * don't show empty partitions
	     */
	    if (qikdata_gendisk.part[minnr].nr_sects == 0)
		continue;

	    if (j == 0)
	    {
		len += sprintf(buf + len, "%s:  ",
		    devExt->UnitDet[i].drive_name);
	    }
	    else
	    {
		len += sprintf(buf + len, "%s%d: ",
		    devExt->UnitDet[i].drive_name, j);
	    }
	    len += sprintf(buf + len, "%10d %10ld %10ld %10d %10d %10d\n",
		qikdata_gendisk.sizes[minnr],
		qikdata_gendisk.part[minnr].start_sect,
		qikdata_gendisk.part[minnr].nr_sects,
		qd_sizes[minnr],
		qd_blocksizes[minnr],
		qd_hardsects[minnr]);
	}
    }
    return (len);
}

/*
 * You can set the QikDATA major device number via the "qikdata_major"
 * module parameter.  By default, the major device number is automatically
 * allocated by the kernel (the first available major device number less
 * than 255, usually 254).
 *
 * Note: for a statically linked driver, this goes on the kernel command
 * line in the /etc/lilo.conf or /etc/grub.conf file.  For this reason, it
 * is important to have a "qikdatad_" wart on the front of parameter
 * names so we don't get collissions with outher parameter names for
 * other modules.
 */
MODULE_PARM(qikdata_major, "i");

/*
 * You can enable or disable the cable ECC checking by using the
 * "qikdata_cable_ecc" parameter.  This is enabled by default; you
 * probably don't want to mess with it, or you risk data corruption.
 */
MODULE_PARM(qikdata_cable_ecc, "i");


#if 0

static void
add_gendisk(struct gendisk *gdp)
{
    gdp->next = gendisk_head;
    gendisk_head = gdp;
}


static void
del_gendisk(struct gendisk *gdp)
{
    struct gendisk **gdpp;

    for (gdpp = &gendisk_head; *gdpp; gdpp = &((*gdpp)->next))
    {
	if (*gdpp == gdp)
	{
	    *gdpp = gdp->next;
	    break;
	}
    }
}

#endif


static int __init
qikdata_load(void)
{
    int     i;
    int     j;
    int     minnr;
    int     majnr;
    int     len;
    int     result;
    int     k = 0;
    PHW_DEVICE_EXTENSION devExt;
    PUNIT_DETAILS unitDet;
    kdev_t  i_rdev;

    printk(KERN_ALERT HERE "load, version %s "
#if CONFIG_SMP
	"SMP"
#else
	"UP"
#endif
	", " __TIME__ " " __DATE__ "\n", qd_version());

    /*
     * Copy the (static) cfg variables to public prefixed ones to allow
     * snoozing with a debugger.
     */

    /*
     * qikdata_major set from module parameters
     */
    qd_devs = devs;
    qd_rahead = rahead;
    qd_blksize = blksize;
    qikdata_hardsect = hardsect;

    /*
     * Hardsect can't be changed :(
     */
    if (hardsect != 512)
    {
	HDEBUG(KERN_ERR HERE "can't change hardsect size\n");
	hardsect = qikdata_hardsect = 512;
    }

    /*
     * Register the block device.
     */
    result = register_blkdev(qikdata_major, DRIVER_LONG_NAME, &qikdata_fops);
    if (result < 0)
    {
	HDEBUG(KERN_ERR HERE "can't get major %d\n", qikdata_major);
	return (result);
    }
    if (qikdata_major == 0)
	qikdata_major = result;	/* dynamic */
    majnr = qikdata_major;		/* Use local majnr or devExt->major */

    qikdata_gendisk.major = majnr;

    /*
     * Assign values common to the MAJOR
     */
#ifndef LINUX_24
    blk_dev[majnr].request_fn = qikdata_request;
#endif
    read_ahead[majnr] = qd_rahead;

    /*
     * Scan the PCI for installed QikDATA cards
     */
    qd_devs = 0;
    qd_devs = qikdata_find_hba(qd_devs, QIKDATA_HBA_CARD);
    if (qd_devs == 0)
    {
	/*
	 * No devices were found, however this *isn't* an error
	 * sufficient to say that the module didn't load successfully.
	 * If the user attempts to open any of our devices, she will be
	 * given an ENODEV (No such device or addres) error.
	 */
	HDEBUG(KERN_ALERT HERE "no devices found\n");
    }
    else
	HDEBUG(KERN_ALERT HERE "Number of devices found = %d\n", qd_devs);

    /*
     * Allocate & clear the arrays of pointers
     */
    memset(qd_sizes, 0, sizeof(qd_sizes));
    memset(qd_partitions, 0, sizeof(qd_partitions));

    qikdata_gendisk.sizes = qd_sizes;
    qikdata_gendisk.part = qd_partitions;

    for (i = 0; i < qd_devs; i++)
    {
	devExt = &(qikdata_device_extension[i]);
	devExt->major = majnr;
	HDEBUG(KERN_ALERT HERE "spin_lock_init(&devExt->lock);\n");
	spin_lock_init(&devExt->lock);

	/*
         * This is used in the /proc/driver/qikdata directory to name
	 * the directory elements.  It is used to refer to the QikDATA
	 * HBA, not the attached QikDATA drives.
	 */
	sprintf(devExt->qd_name, "qb%c", 'a' + i);

	/*
	 * don't process requests during initialsation
	 */
	devExt->PendingRequest = PENDING_REQUEST_BOGUS;

	/*
	 * Request interrupts here
	 * - this really should be done at open time...
	 */
	minnr = QD_MK_MIN(i, 0, 0);	/* HBA's minor dev num */
	result =
	    request_irq(devExt->irq, HwInterrupt, SA_INTERRUPT | SA_SHIRQ,
	    DRIVER_LONG_NAME, (void *)MKDEV(devExt->major, minnr));
	if (result < 0)
	{
	    HDEBUG(KERN_ERR HERE "can't get assigned IRQ %d\n",
		(int)(devExt->irq));

	    /*
	     * FUTURE: check if cleaning up properly
	     */
	    goto fail_malloc;
	}

	/*
	 * allocate a page for SGLs, ECC and BANKs etc
	 * FUTURE: check if needed
	 */
	devExt->NoncachedExtension = (void *)__get_free_page(GFP_DMA);
	if (!devExt->NoncachedExtension)
	{
	    result = -ENOMEM;
	    goto fail_malloc;
	}
	devExt->eccBufferAddr =
	    virt_to_bus(devExt->NoncachedExtension->eccBuffer);
	devExt->bankBufferAddr =
	    virt_to_bus(devExt->NoncachedExtension->bankBuffer);

	/*
	 * allocate a page for DMA intermediate buffer
	 * - for use in format etc
	 */
	devExt->ioBuffer = (void *)__get_free_page(GFP_DMA);
	if (!devExt->ioBuffer)
	{
	    result = -ENOMEM;
	    goto fail_malloc;
	}
	devExt->ioBufferAddr = virt_to_bus(devExt->ioBuffer);

	/*
	 * Initialise the qd hardware
	 * device num to access all drives on the HBA
	 */
	i_rdev = MKDEV(devExt->major, QD_MK_MIN(i, 0, 0));

	/*
	 * don't process requests during initialsation
	 */
	devExt->PendingRequest = PENDING_REQUEST_BOGUS;
	HwInitialise(i_rdev);
	devExt->PendingRequest = 0;

	/*
	 * Wait for the device initialisation to complete
	 * we need to do this because the RRPART ioctl call
	 * is going to start accessing the disk
	 */
	while (!devExt->DeviceReady)
	{
	    if (k++ >= MAX_TRIES)
	    {
		/*
		 * just break for now, might need code
		 * to clear malloced memory etc
		 */
		break;
	    }
	}
	devExt->UnitIndex = 0;	/* used for single drive driver */
	unitDet = &(devExt->UnitDet[devExt->UnitIndex]);

	if (devExt->DeviceReady)
	{
#ifdef LINUX_24
	    request_queue_t *q;

	    q = BLK_DEFAULT_QUEUE(devExt->major);
	    q->queuedata = devExt;
	    blk_init_queue(q, qikdata_request);
	    blk_queue_headactive(q, 0);
#endif

	    for (j = 0; j < devExt->nBays; j++)
	    {
		unsigned long total_sectors;
		int     drv_minr;

		drv_minr = QD_MK_MIN(i, j, 0);	/* drive's minor dev num */
		unitDet = &(devExt->UnitDet[j]);

		/*
		 * still using 512-byte blocks for the moment
		 * and qd_sizes wants number of kilobytes
		 * - we deferred this from earlier on...
		 *
		 * We are going to answer the geometry request in multiples
		 * of FAKE_SECTORS*FAKE_HEADS, so round down to match the
		 * answer we are going to give.
		 */
		total_sectors = unitDet->Dimm[MAX_DIMMS - 1].totalBlocks;
		total_sectors -= total_sectors % (FAKE_HEADS * FAKE_SECTORS);
		qd_sizes[drv_minr] = total_sectors / 2;

		/*
		 * fill whole-disk entries
		 * start sector is already zero, and sects are 512 bytes long
		 */
		qd_partitions[drv_minr].nr_sects = qd_sizes[drv_minr] * 2;
		qd_blocksizes[drv_minr] = qd_blksize;
		qd_hardsects[drv_minr] = qikdata_hardsect;
	    }

	    /*
	     * Initialise driver statistics
	     */
	    devExt->LargestTransferSize = 0;
	    devExt->ReceivedUnaligned = 0;

	    /*
	     * store time the driver started
	     */
	    devExt->StartTime = xtime;
	}
    }
    qikdata_gendisk.nr_real = qd_devs;

    /*
     * now point to our malloced arrays
     */
    blk_size[majnr] = qd_sizes;
    blksize_size[majnr] = qd_blocksizes;
#ifdef USE_HARDSECTS
    hardsect_size[majnr] = qd_hardsects;
#endif

    for (i = 0; i < qd_devs; i++)
    {
	devExt = &(qikdata_device_extension[i]);
	unitDet = &(devExt->UnitDet[devExt->UnitIndex]);

	/*
	 * device num to access all drives on the HBA
	 */
	minnr = QD_MK_MIN(i, 0, 0);	/* HBA's minor dev num */
	i_rdev = MKDEV(devExt->major, minnr);
	qikdata_proc_register(i_rdev);

	/*
	 * Might need some delay loop checking DeviceReady if initialisation
	 * is non_blocking.
	 */
	if (devExt->DeviceReady)
	{
	    for (j = 0; j < devExt->nBays; j++)
	    {
#ifdef LINUX_24
		kdev_t  drv_rdev;
		int     drv_minr;

		drv_minr = QD_MK_MIN(i, j, 0);
		drv_rdev = MKDEV(devExt->major, drv_minr);
		unitDet = &(devExt->UnitDet[j]);
		HDEBUG(KERN_ALERT HERE "Initial Partition Table Setup:\n");
		qikdata_dump_gendisk(&qikdata_gendisk, drv_rdev, QD_MAXRDEV,
		    (void *)&qikdata_fops, qd_partitions[drv_minr].nr_sects);
		len = qikdata_dump_partition_table(drv_rdev, dump_buff);
		dump_buff[len] = 0;
		HDEBUG(KERN_ALERT HERE "%s\n", dump_buff);
		HDEBUG(KERN_ALERT HERE "register_disk {\n");
		register_disk(&qikdata_gendisk, drv_rdev, QD_MAXRDEV,
		    (void *)&qikdata_fops, qd_partitions[drv_minr].nr_sects);
		HDEBUG(KERN_ALERT HERE "register_disk }\n");
#else
		HDEBUG(KERN_ALERT HERE "resetup_one_dev {\n");
		resetup_one_dev(&qikdata_gendisk, i);
		HDEBUG(KERN_ALERT HERE "resetup_one_dev }\n");
#endif
	    }
	}
	else
	{
	    printk(KERN_ALERT HERE "Device not ready at end of init. loop.\n");
	}
    }

#ifdef _INSERT_GENDISK_IN_LIST
    /*
     * insert our gendisk structure into the kernel list
     */
    add_gendisk(&qikdata_gendisk);
#endif

    return (0);			/* succeed */

  fail_malloc:
    read_ahead[majnr] = 0;
    blk_size[majnr] = NULL;
    blksize_size[majnr] = NULL;
    hardsect_size[majnr] = NULL;

    /*
     * ### fix this!!!
     */
    if (devExt->NoncachedExtension)
	free_page((u_long) devExt->NoncachedExtension);
    if (devExt->ioBuffer)
	free_page((u_long) devExt->ioBuffer);

    unregister_blkdev(majnr, DRIVER_LONG_NAME);

    return (result);
}

#ifndef __exit
#define __exit
#endif

static void __exit
qikdata_unload(void)
{
    PHW_DEVICE_EXTENSION devExt;
    u_int   pRTR;
    u_int   pMEM;
    PAGE_REG pageRegister;
    int     i;
    int     j;
    u32     dma;
    kdev_t  i_rdev;
    int     majnr;
    int     minnr;

    printk(KERN_ALERT HERE "unload, version %s "
#if CONFIG_SMP
	"SMP"
#else
	"UP"
#endif
	", " __TIME__ " " __DATE__ "\n", qd_version());

    /*
     * first of all, flush it all and reset all the data structures
     */
#ifdef _INSERT_GENDISK_IN_LIST
    /*
     * remove qikdata_gendisk from the kernel list
     * FUTURE: put gendisk into devExt structure
     */
    del_gendisk(&qikdata_gendisk);
#endif
    for (i = 0; i < qd_devs; i++)
    {
	unsigned long flags;

	devExt = &(qikdata_device_extension[i]);
	pRTR = (u_int) devExt->RTRmap;
	pMEM = (u_int) devExt->LASmap;
	majnr = devExt->major;

	/*
	 * flush all partitions. Loops through all major/minor
	 * combinations.
	 * This loop is for HBAs, FUTURE: add loop for drives
	 */
	for (j = 0; j << (1 << QD_SHIFT); j++)
	{
	    minnr = QD_MK_MIN(i, 0, j);
	    i_rdev = MKDEV(majnr, minnr);
	    fsync_dev(i_rdev);
	}
#ifndef LINUX_24
	blk_dev[majnr].request_fn = NULL;
#endif
	read_ahead[majnr] = 0;
	blk_size[majnr] = NULL;
	blksize_size[majnr] = NULL;
	hardsect_size[majnr] = NULL;

	/*
	 * Signal to the drives that we are unloading
	 */
	for (j = 0; j < devExt->nBays; j++)
	{
	    PUNIT_DETAILS unitDet = &(devExt->UnitDet[j]);

	    HDEBUG(KERN_ALERT HERE "signal QikDATA Drive %d we are unloading\n",
		unitDet->drive_address);
	    HDEBUG(KERN_ALERT HERE "spin_lock_irqsave(&devExt->lock);\n");
	    spin_lock_irqsave(&devExt->lock, flags);
	    pageRegister.dw = (PAGE_MASK_NIOS | PAGE_MASK_EXTENDED);
	    pageRegister.s.drive = unitDet->drive_address;
	    writel(pageRegister.dw, pRTR + REG_PAGE);
	    writel(0, pMEM + CABLE_WE_DRVLOAD);
	    writel(0, pMEM + CABLE_WE_DRVLOAD + 4);
	    udelay(20);
	    writel(0, pRTR + REG_PAGE);
	    HDEBUG(KERN_ALERT HERE "spin_unlock_irqrestore(&devExt->lock);\n");
	    spin_unlock_irqrestore(&devExt->lock, flags);
	}

	/*
	 * Reset hardware and release resource mappings
	 */
	HwReset(i);

	/*
	 * Reset the Acex register
	 */
	devExt->AcexIOReg = 0;
	writel(devExt->AcexIOReg, pRTR + REG_IO);

	/*
	 * turn on ATTN LED
	 */
	devExt->AcexIOReg |= IO_nLED_ATTN;
	writel(devExt->AcexIOReg, pRTR + REG_IO);

	/*
	 * tell if any interrupts still pending
	 */
	dma = readl(pRTR + REG_DMA);
	if (dma & DMA_INTERRUPT_PENDING)
	    printk(KERN_ALERT HERE "Interrupt still pending at cleanup\n");

	if (dma & DMA_PCI_CORE_INTERRUPT)
	{
	    ushort  pciISR;

	    printk(KERN_ALERT HERE "PCIAbortInterruptActive\n");
	    /*
	     * Clear all possible PCI-related interrupts
	     */
	    pciISR =
		(PCI_STATUS_DATA_PARITY_DETECTED |
		PCI_STATUS_SIGNALED_TARGET_ABORT |
		PCI_STATUS_RECEIVED_TARGET_ABORT |
		PCI_STATUS_RECEIVED_MASTER_ABORT |
		PCI_STATUS_SIGNALED_SYSTEM_ERROR |
		PCI_STATUS_DETECTED_PARITY_ERROR);
	    pci_write_config_word(devExt->pci_dev, REG_PCIISR, pciISR);
	}

	/*
	 * clear all possible interrupts
	 */
	HDEBUG(ISR_DEBUG HERE "dma = %s (0x%08X)\n", dma_decode(dma), dma);
	writel(dma | DMA_SGL_INTERRUPT | DMA_ABORT_INTERRUPT |
	    DMA_PCI_CORE_INTERRUPT | DMA_ECC_INTERRUPT, pRTR + REG_DMA);

	/*
	 * device num to access the whole HBA/drive set
	 */
	i_rdev = MKDEV(majnr, QD_MK_MIN(i, 0, 0));
	free_irq(devExt->irq, (void *)((ulong) i_rdev));

	/*
	 * free the virtual address mapping to the memory space
	 */
	iounmap(devExt->LASmap);
	iounmap(devExt->RTRmap);

	/*
	 * free the dma buffers
	 */
	free_page((u_long) devExt->NoncachedExtension);
	free_page((u_long) devExt->ioBuffer);

	qikdata_proc_unregister(i_rdev);

	/*
	 * Unregister the block device.
	 */
	unregister_blkdev(majnr, DRIVER_LONG_NAME);
    }
    qikdata_proc_unregister_dir();
}

module_init(qikdata_load);
module_exit(qikdata_unload);

MODULE_AUTHOR("Platypus Technology Holdings, Inc.");
#ifdef MODULE_LICENSE
MODULE_LICENSE("BSD");
#endif
