/*
 * drivers.h - cable space access routines
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#ifndef FIRMWARE_MSW_SRC_DRIVERS_H
#define FIRMWARE_MSW_SRC_DRIVERS_H

#undef DEBUG_CABLE_ACCESS

/*
 * Cable access addresses.  Each register occupies 8 bytes of address
 * space, so we number them in increments of 8.
 *
 * Notes about cable registers:
 *
 * 1. All of the values that are reported are updated as they change,
 * so you can poll for changes.
 *
 * 2. All packs return 0 for unimplemented cable register reads.
 *
 * 3. Unimplemented cable register writes are ignored.
 */

/* Extended, dimm space */

/**
 * CABLE_ED_SPDS - per-DIMM SPDs.  This returns the geometry of a
 * DIMM.  The cable register command coming down should have a value
 * based on CABLE_SGL from qif_hw.h, with the "nios",
 * "dimm_bits_valid" and "extended" bits set to 1, and the "dimm"
 * field set to the relevent DIMM.
 *
 * Value: bits 0..31 = a DIMM geometry, as defined in DIMM_BASIS from sdram.h
 *        bits 32..63 = the DIMM serial number
 *
 * Notes: present in pack 16
 */
#define CABLE_ED_SPDS		(0<<3)	/* per-dimm spds */


/* Extended, non-dimm space */

/**
 * CABLE_E_MAGIC - magic number and serial number.
 *
 * Value: bits 0..31 = CABLE_MAGIC_QBAY, ("QBAY")
 *        bits 32..63 = the serial number of the bay.
 *
 * Notes: present in pack 16
 */
#define CABLE_E_MAGIC		(0<<3)	/* magic number and serial */
#define CABLE_MAGIC_QBAY        0x59414251	/* "QBAY" */

/**
 * CABLE_WE_DRVLOAD - signal driver load
 *
 * Values: bits 0..63 = a write to this will inform the bay that
 *                      the driver has attached (value = non-zero) or
 *                      detached (value = 0).  Currently any write to
 *                      this will have the effect of resetting the ECC
 *                      statistics.
 *
 * Notes: introduced in pack 19
 */
#define CABLE_WE_DRVLOAD	(0<<3)	/* signal driver load */

/**
 * CABLE_E_FIRMVER - firmware/ups versions.
 *
 * Value: bits 0..15 = FPGA version
 *        bits 16..31 = CPLD version
 *        bits 32..47 = NIOS monitor version
 *        bits 48..63 = UPS micro version
 *
 * Notes: present in pack 16
 */
#define CABLE_E_FIRMVER		(1<<3)	/* firmware/ups versions */

/**
 * CABLE_E_MSWVER - msw version string
 *
 * Value: a character string with length 8 bytes, without a null
 *        termination.  (In pack 19 and later, it will be similar to
 *        "1.1.D286", while in pack 18 and earlier, it will be similar
 *        to "@G-1")
 *
 * Notes: present in pack 16
 */
#define CABLE_E_MSWVER		(2<<3)	/* msw version string */

/**
 * CABLE_E_MSWBUILD - msw build string
 *
 * Value: a character string with length up to 8 bytes.  (In pack 19
 *        and later, it will usually be "1", and in pack 18 and earlier, it
 *        will be an arbitrary character string of decomal digits.)
 *
 * Notes: present in pack 16
 */
#define CABLE_E_MSWBUILD	(3<<3)	/* msw build string */

/**
 * CABLE_E_ERRSTATUS - error status
 *
 * Value: bits 0..31 = current active alerts in the bay.  The bit
 *                     definitions are in the defs file dfi_alert.h
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_ERRSTATUS	(4<<3)	/* error status */

/**
 * CABLE_E_RESVBLOCKS - reserved sdram blocks
 *
 * Value: bits 0..31 = the number of 512-byte blocks reserved for the
 *                      private use by the bay.  Currently 0.
 *        bits 32..63 = unused
 *
 * Notes: present in pack 16
 */
#define CABLE_E_RESVBLOCKS	(5<<3)	/* reserved sdram blocks */

/**
 * CABLE_WE_RESVBLOCKS - driver reserved sdram blocks
 *
 * Value: bits 0..31 = the number of 512-byte blocks the driver wishes
 *                     to reserve.
 *        bits 32..63 = unused
 *
 * Notes: present in pack 16
 */
#define CABLE_WE_RESVBLOCKS	(5<<3)	/* driver reserved sdram blocks */

/**
 * CABLE_E_PACK_VERSION - pack version number
 *
 * Value: bits 0..15 = the pack version number, in binary form.  Eg. 19,
 *                     20, etc if the firmware is a consistent pack release,
 *                     0 if the firmware is pre-pack 19, or not a
 *                     consistent released pack
 *        bits 16..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_PACK_VERSION    (6<<3)  /* pack version number */

/**
 * CABLE_E_EARLIEST_COMPATIBLE_PACK - earliest compatible pack version number
 *
 * Value: bits 0..15 = the earliest compatible pack version number, in
 *                     binary form.  So if the firmware is pack 20, and it is
 *                     backward compatible with pack 19, then the
 *                     value will be 19.
 *        bits 16..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_EARLIEST_COMPATIBLE_PACK (7<<3) /* earliest compatible pack
					   version number */

/**
 * CABLE_E_TIME - time in seconds since 1/1/1970.  This is settable
 * and gettable.  If you set this, all of the times you retrieve from
 * the bay, (such as backup timestamp) will be the "right time".  Also,
 * the log file in flash ram will have events stamped with the right time.
 *
 * Value: bits 0..31 = the time in seconds since midnight before 1/1/1970.
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_TIME            (8<<3)  /* time in seconds since 1/1/1970 */

/**
 * CABLE_WE_BUTTONS - arbitrary button press for debug.  This allows
 * the host to simulate the pressing of buttons on the bay, which will
 * drive the bay around its menuing system.
 *
 * Value: bits 0..3 = the button presses (defined in dfi_buttons.h)
 *
 * Notes: introduced in pack 19
 */
#define CABLE_WE_BUTTONS        (9<<3)  /* arbitrary button press for debug */

/* Extended space, read-write */


/**
 * CABLE_RW_SDRAM_CORR - SDRAM correctable errors.  This is a readable
 * and writeable register.  Reads return the current value of
 * accumulated errors.  Writes clear that count.  Separate counts are
 * kept for the LCD error count display and the cable register
 * interface count, so if the host clears this count, it will not
 * affect the LCD display count.
 *
 * Value: bits 0..31 = the error count
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_RW_SDRAM_CORR     (10<<3) /* SDRAM correctable errors */

/**
 * CABLE_RW_SDRAM_UNCORR - SDRAM uncorrectable errors.  This is a readable
 * and writeable register.  Reads return the current value of
 * accumulated errors.  Writes clear that count.  Separate counts are
 * kept for the LCD error count display and the cable register
 * interface count, so if the host clears this count, it will not
 * affect the LCD display count.
 *
 * Value: bits 0..31 = the error count
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_RW_SDRAM_UNCORR   (11<<3) /* SDRAM uncorrectable errors */

/**
 * CABLE_RW_LVDS_CORR - LVDS correctable errors.  This is a readable
 * and writeable register.  Reads return the current value of
 * accumulated errors.  Writes clear that count.  Separate counts are
 * kept for the LCD error count display and the cable register
 * interface count, so if the host clears this count, it will not
 * affect the LCD display count.  This only counts *incoming* LVDS
 * errors, ie. from the previous bay or QikInterface.
 *
 * Value: bits 0..31 = the error count
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_RW_LVDS_CORR      (12<<3) /* LVDS correctable errors */

/**
 * CABLE_RW_LVDS_UNCORR - LVDS uncorrectable errors.  This is a readable
 * and writeable register.  Reads return the current value of
 * accumulated errors.  Writes clear that count.  Separate counts are
 * kept for the LCD error count display and the cable register
 * interface count, so if the host clears this count, it will not
 * affect the LCD display count.  This only counts *incoming* LVDS
 * errors, ie. from the previous bay or QikInterface.
 *
 * Value: bits 0..31 = the error count
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_RW_LVDS_UNCORR    (13<<3) /* LVDS uncorrectable errors */


/**
 * CABLE_RW_TEMPERATURE_THRESH - set/get upper temperature alarm
 * threshold.  This is 50 degrees celsius by default.  Setting this
 * will update the value stored in EEPROM, so the new value will
 * survive reboots.
 *
 * Value: bits 0..7 = the temperature threshold in degrees celsius
 *        bits 8..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_RW_TEMPERATURE_THRESH (14<<3)

/**
 * CABLE_E_BATT_PRESENT - ups battery present
 *
 * Value: bits 0..15 = battery 1 present (0 = not present, 1 = present)
 *        bits 16..31 = battery 2 present (0 = not present, 1 = present)
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_BATT_PRESENT    (15<<3)  /* ups battery present */

/**
 * CABLE_E_BATT_TEST_PROG - battery test progress
 *
 * Value: bits 0..15 = 1 if battery 1 test has completed, 2 if battery
 *                     2 test has completed, otherwise 0 if no battery
 *                     tests have been run.
 *        bits 16..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_BATT_TEST_PROG  (16<<3)  /* battery test progress */

/**
 * CABLE_E_BATT_VOLTAGE - battery voltage
 *
 * Value: bits 0..15 = battery 1 voltage (units = 1/10ths of a Volt)
 *        bits 16..31 = battery 2 voltage (units = 1/10ths of a Volt)
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_BATT_VOLTAGE    (17<<3) /* battery voltage */

/**
 * CABLE_RW_BATT_VOLT_THRESHOLD - battery voltage alarm threshold.  Setting
 * this will update the value stored in EEPROM, so the new value will
 * survive reboots.
 *
 * Value: bits 0..15 = the voltage threshold, in units of 1/10ths of a Volt
 *        bits 16..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_RW_BATT_VOLT_THRESHOLD (18<<3) /* battery voltage threshold */

/**
 * CABLE_E_HDD_PRESENT - whether the hard disks are present
 *
 * Value: bits 0..15 = HDD 1 exists (1 = exists, 0 = doesn't exist)
 *        bits 16..31 = HDD 2 exists (1 = exists, 0 = doesn't exist)
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_HDD_PRESENT     (19<<3) /* whether the hard disks are present */

/**
 * CABLE_E_HDD_1_SIZE - size of hard disk 1 (in blocks)
 *
 * Value: bits 0..31 = size of HDD 1 in 512-byte blocks
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_HDD_1_SIZE      (20<<3) /* size of hard disk 1 (in blocks) */

/**
 * CABLE_E_HDD_2_SIZE - size of hard disk 2 (in blocks)
 *
 * Value: bits 0..31 = size of HDD 2 in 512-byte blocks
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_HDD_2_SIZE      (21<<3) /* size of hard disk 2 (in blocks) */

/**
 * CABLE_E_BACKUP_STATUS - show the status of the backups
 *
 * Values: bits 0..15 = HDD 1 backup status
 *         bits 16..31 = HDD 2 backup status
 *         bits 32..63 = unused
 *
 * The backup status values are defined in dfi_backup.h
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_BACKUP_STATUS   (22<<3) /* show the status of the backups */

/**
 * CABLE_E_LAST_RESTORE_HDD - drive last restore was from
 *
 * Value: bits 0..7 = which drive provided the last restore.  (0 for
 *                    drive A: and 1 for drive B:)
 *        bits 8..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_LAST_RESTORE_HDD (23<<3) /* drive last restore was from */

/**
 * CABLE_E_NUM_OF_REBOOTS - number of QikData reboots
 *
 * Value: bits 0..31 = the number of times this bay has been powered
 *                     up.
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_NUM_OF_REBOOTS  (24<<3) /* number of QikData reboots */

/**
 * CABLE_E_BACKUP_1_SIZE - size of backup on 1st hdd (blocks)
 *
 * Value: bits 0..31 = the backup size of HDD 1 in 512-byte blocks.
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_BACKUP_1_SIZE   (25<<3) /* size of backup on 1st hdd (blocks) */

/**
 * CABLE_E_BACKUP_2_SIZE - size of backup on 2nd hdd (blocks)
 *
 * Value: bits 0..31 = the backup size of HDD 2 in 512-byte blocks.
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_BACKUP_2_SIZE   (26<<3) /* size of backup on 2nd hdd (blocks) */

/**
 * CABLE_E_BACKUP_1_TIMESTAMP - a timestamp for the start of the
 * backup on HDD 1.  This will be the time in seconds since 1/1/1970,
 * as long as the CABLE_E_TIME cable register has been initialised.
 * Otherwise it will be time in seconds since the previous NIOS boot.
 * This is a read-only register.
 *
 * Value: bits 0..31 = the backup timestamp of HDD 1
 *        bits 32..63 = unused
 *
 * Notes: introduced in X2v3
 */
#define CABLE_E_BACKUP_1_TIMESTAMP (27<<3) /* backup 1 timestamp */

/**
 * CABLE_E_BACKUP_2_TIMESTAMP - a timestamp for the start of the
 * backup on HDD 2.  This will be the time in seconds since 1/1/1970,
 * as long as the CABLE_E_TIME cable register has been initialised.
 * Otherwise it will be time in seconds since the previous NIOS boot.
 * This is a read-only register.
 *
 * Value: bits 0..31 = the backup timestamp of HDD 2
 *        bits 32..63 = unused
 *
 * Notes: introduced in X2v3
 */
#define CABLE_E_BACKUP_2_TIMESTAMP (28<<3) /* backup 2 timestamp */

/**
 * CABLE_E_LOAD_NEW_FPGA - loads a new fpga from an arbitary 0.5 MB block of
 * flash. The address in flash is written through this cable register, where
 * the address is shifted down by 19 i.e. 0 for the first block, 1 for the
 * second, 3 for the third, etc. So far these FPGAs are defined:
 * Block 0 - normal FPGA
 * Block 1 - failsafe FPGA
 * Block 3 - factory test firmware FPGA
 * Block 4 - debug FPGA
 * Reading this register will be 1 if this CPLD supports reloading the FPGA
 * and 0 otherwise.
 * This is a read/write register.
 *
 * Value: bits 0..3 = 0.5 MB block of flash to load the FPGA from (writing)
 *        bit  0    = 1 if this CPLD supports it and 0 otherwise (reading)
 *        bits 4..63 = unused
 *
 * Notes: introduced in X2v4
 */
#define CABLE_E_LOAD_NEW_FPGA (29<<3) /* load new fpga */


/* spare: 30, 31 */

/**
 * CABLE_E_BATTCAP - ups battery capacity
 *
 * Value: bits 0..15 = battery 1 capacity (measured in 1/10ths of
 *                     Ampere-hours, or 999 if unknown)
 *        bits 16..31 = battery 2 capacity (measured in 1/10ths of
 *                     Ampere-hours, or 999 if unknown)
 *        bits 32..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_BATTCAP		(32<<3)	/* ups battery capacity */

/**
 * CABLE_E_TEMP - temperature
 *
 * Value: bits 0..15 = UPS board temperature (degrees celsius)
 *        bits 16..31 = DIMM temperature (degrees celsius)
 *        bits 32..47 = FPGA temperature (degrees celsius)
 *        bits 48..63 = Power supply temperature (degrees celsius)
 *
 * Notes: present in pack 16
 */
#define CABLE_E_TEMP		(33<<3)	/* temperature */

/**
 * CABLE_E_FANSPEED - fan speed
 *
 * Value: bits 0..15 = fan 1 speed (revolutions per minute)
 *        bits 16..31 = fan 2 speed (revolutions per minute)
 *        bits 32..47 = fan 3 speed (revolutions per minute)
 *        bits 48..63 = fan 4 speed (revolutions per minute)
 *
 * Notes: present in pack 16
 */
#define CABLE_E_FANSPEED	(34<<3)	/* fan speed */

/**
 * CABLE_E_LCD_TOP_LEFT, CABLE_E_LCD_TOP_RIGHT,
 * CABLE_E_LCD_BOTTOM_LEFT, CABLE_E_LCD_BOTTOM_RIGHT - ask the NIOS
 * what is being displayed on the LCD screen.  There are 2 lines of 16
 * characters each, so you need 4 cable registers to get all 32
 * characters.  Each cable read returns 8 characters, from the top or
 * bottom rown, left or right half.
 *
 * Value: bits 0..63 = eight characters being displayed on the LCD
 *
 * Notes: introduced in X2v3
 */
#define CABLE_E_LCD_TOP_LEFT      (35<<3)
#define CABLE_E_LCD_TOP_RIGHT     (36<<3)
#define CABLE_E_LCD_BOTTOM_LEFT   (37<<3)
#define CABLE_E_LCD_BOTTOM_RIGHT  (38<<3)

/**
 * CABLE_WR_FLASH_WRITE_ADDR_LEN - A write to this register will set
 * up the address and length in quadwords which will be used for
 * modifying the flash RAM.  It will also set up a "signature" value
 * which must be set correctly before any of the "flash operation"
 * actions will be attempted.  These value should be set up before
 * using the next cable registers to write to flash.
 *
 * A read from this register will return the values previously set,
 * though the address may be updated by some of the flash operations.
 *
 * The actual writing to flash RAM is done by writes to the cable regs
 * CABLE_WR_FLASH_WRITE_SDRAM and CABLE_WR_FLASH_WRITE_QWORD.  When
 * they are written to to update flash RAM, the "flash write address"
 * is updated to point to the next unwritten piece of flash RAM.
 *
 * The address must be quadword aligned.  The length is in quadwords.
 * That is, you can only transfer multiples of 8 bytes, and addresses
 * must be divisible by 8.
 *
 * After the flash write has finished, the address is updated to the
 * next place in flash to be written.
 *
 * Since the flash modification cable register operations are
 * inherently dangerous, it is a requirement that there is a signature
 * called FLASH_WRITE_SIGNATURE in the top 16 bits of the cable write.
 * This value is defined in dfi_flash.h.  If this signature is not
 * set, no flash modification operations will be attempted.
 *
 * Value: bits 0..31 = the 32 bit address in flash RAM that the
 *                     following flash writes should be written to
 *        bits 32..47 = the length in quadwords to be written to flash RAM
 *        bits 48..63 = FLASH_WRITE_SIGNATURE (as defined in dfi_flash.h)
 *
 * Notes: introduced in X2v4
 */
#define CABLE_WR_FLASH_WRITE_ADDR_LEN (39<<3)

/**
 * CABLE_WR_FLASH_WRITE_SDRAM - A write to this cable register will
 * trigger a write of a block of memory from SDRAM to flash RAM.  The
 * flash RAM address to be written to is specified in the bottom 32
 * bits of the CABLE_WR_FLASH_WRITE_ADDR_LEN register.  The block size
 * to be copied is specified in the CABLE_WR_FLASH_WRITE_ADDR_LEN
 * register.  The address in SDRAM that the block will be copied from
 * is specified by the SDRAM vector which is passed down in this cable
 * write.  The address specified in CABLE_WR_FLASH_WRITE_ADDR_LEN
 * register will be incremented by the block size.  A read from this
 * register will be non-zero if there is a pending flash operation,
 * and 0 if there is not.
 *
 * The flash address, block size and flash operation signature must
 * have been set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN
 * cable write, or this won't work.
 *
 * Value: bits 0..63 = an SDRAM vector specifying what to write to flash (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_WR_FLASH_WRITE_SDRAM (40<<3)

/**
 * CABLE_WR_FLASH_WRITE_QWORD - A write to this cable register will
 * trigger a write of the 8 bytes to flash RAM.  The 8 bytes will be
 * written to the address specified in CABLE_WR_FLASH_WRITE_ADDR_LEN
 * register.  That address will be incremented by 8.  A read from this
 * register will be non-zero if there is a pending flash write or
 * erasure, and 0 if there is not.
 *
 * The flash address and flash operation signature must have been
 * set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN cable write,
 * or this won't work.
 *
 * Value: bits 0..63 = the 8 bytes to write to flash (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_WR_FLASH_WRITE_QWORD (41<<3)

/**
 * CABLE_WR_FLASH_ERASE - a write to this cable register will trigger
 * an erasure of a block of flash RAM.  The block to be erased will be
 * the one specified in the CABLE_WR_FLASH_WRITE_ADDR_LEN cable
 * register.  The block size erased will always be the block size of
 * the flash RAM, which is 128kB.  A read from this register will be
 * non-zero if there is a pending flash operation, and 0 if there
 * is not.
 *
 * The flash address and flash operation signature must have been
 * set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN cable write,
 * or this won't work.
 *
 * Value: bits 0..63 = any value (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_WR_FLASH_ERASE (42<<3)

/**
 * CABLE_WR_FLASH_LOCK - a write to this cable register will trigger
 * the locking of a flash RAM block.  The block to be locked will be
 * the one specified in the CABLE_WR_FLASH_WRITE_ADDR_LEN cable
 * register.  The block size locked will always be the block size of
 * the flash RAM, which is 128kB.  A read from this register will be
 * non-zero if there is a pending flash operation, and 0 if there
 * is not.
 *
 * The flash address and flash operation signature must have been
 * set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN cable write,
 * or this won't work.
 *
 * Value: bits 0..63 = any value (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_WR_FLASH_LOCK (43<<3)

/**
 * CABLE_WR_FLASH_UNLOCK - a write to this cable register will trigger
 * the unlocking of the entire flash RAM. A read from this register
 * will be non-zero if there is a pending flash operation, and 0 if
 * there is not.
 *
 * The flash address and flash operation signature must have been
 * set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN cable write,
 * or this won't work.
 *
 * Value: bits 0..63 = any value (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_WR_FLASH_UNLOCK (44<<3)

/**
 * CABLE_WR_FLASH_CHECKSUM - a write to this cable register will
 * trigger the summing of a block of flash RAM.  The address and
 * length (in quadwords) of flash RAM to sum are set up by a write to
 * the CABLE_WR_FLASH_WRITE_ADDR_LEN register.
 *
 * A read from this register will return a 16 bit sum of the region of
 * flash, where the flash RAM is treated as a series of unsigned 16
 * bit integers.  It will also return a "valid" flag, which says
 * whether the sum is correct.  The valid flag is cleared when the
 * checksum is requested, and set when the checksum is finished.
 *
 * The flash address and flash operation signature must have been
 * set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN cable write,
 * or this won't work.
 *
 * Value: bits 0..63 = any value (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *        bits 16..31 = the checksum (read)
 *        bits 32..47 = the CRC (read)
 *
 * Notes: Introduced in X2v4
 *        The CRC is a fairly basic one:
 *            CRC = 0;
 *            for each 16 bit word, CRC = (CRC << 1) XOR word
 */
#define CABLE_WR_FLASH_CHECKSUM (45<<3)

/**
 * CABLE_WR_BOOT_NEW_PROGRAMME - a write to this cable register will
 * trigger the booting of a new NIOS programme in the flash RAM.  A
 * read from this register will be non-zero if there is a pending
 * flash operation, and 0 if there is not.
 *
 * The flash address and flash operation signature must have been
 * set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN cable write,
 * or this won't work.
 *
 * Value: bits 0..63 = any value (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_WR_BOOT_NEW_PROGRAMME (46<<3)

/**
 * CABLE_R_PROGRAMME_VERSION - a read from this cable register will
 * tell the host interesting things about this version's ability to
 * download firmware.
 *
 * Value: bits 0..15 = which 128kB page flash the programme is executing in
 *        bits 16..31 = 1 if this firmware version supports firmware
 *             download, else 0.
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_R_PROGRAMME_VERSION   (47<<3)

/**
 * CABLE_WR_FLASH_READ_SDRAM - A write to this cable register will
 * trigger a read of a block of memory from flash RAM to SDRAM.  The
 * flash RAM address to be written to is specified in the bottom 32
 * bits of the CABLE_WR_FLASH_WRITE_ADDR_LEN register.  The block size
 * to be copied is specified in the CABLE_WR_FLASH_WRITE_ADDR_LEN
 * register.  The address in SDRAM that the block will be copied to
 * is specified by the SDRAM vector which is passed down in this cable
 * write.  The address specified in CABLE_WR_FLASH_WRITE_ADDR_LEN
 * register will be incremented by the block size.  A read from this
 * register will be non-zero if there is a pending flash operation,
 * and 0 if there is not.
 *
 * The flash address, block size and flash operation signature must
 * have been set up correctly by the CABLE_WR_FLASH_WRITE_ADDR_LEN
 * cable write, or this won't work.
 *
 * Value: bits 0..63 = an SDRAM vector specifying what to read from flash
 *                     (write)
 *        bits 0..15 = 0 => no pending flash operation, non-zero => pending
 *                     flash operation (values defined in dfi_flash.h) (read)
 *
 * Notes: Introduced in X2v4
 */
#define CABLE_WR_FLASH_READ_SDRAM (48<<3)

/* spare: 49..63 */

/**
 * CABLE_E_SDRWCOUNT - sdram r/w counters
 *
 * Value: bits 0..63 = unsupported, unused
 *
 * Notes: not supported in any pack so far
 */
#define CABLE_E_SDRWCOUNT	(64<<3)	/* sdram r/w counters */

/**
 * CABLE_E_IDERWCOUNT - ide r/w counters
 *
 * Value: bits 0..31 = number of 512-byte blocks read from a disk
 *                     drive (summ of HDD 1 and HDD 2)
 *        bits 32..63 = number of 512-byte blocks written to a disk
 *                     drive (summ of HDD 1 and HDD 2)
 *
 * Notes: present in pack 16
 */
#define CABLE_E_IDERWCOUNT	(65<<3)	/* ide r/w counters */

/* spare: 66..127 */

/**
 * CABLE_E_NTREADCAP - NT driver support (READ CAPACITY)
 *
 * Value: bits 0..31 = block number of the bay's last block address
 *                     (Note: big-endian!)
 *        bits 32..63 = block size = 512 (Note: big-endian!)
 *
 * Notes: present in pack 16
 */
#define CABLE_E_NTREADCAP	(128<<3)/* NT driver support (READ CAPACITY) */

/* spare: 129 */

/**
 * CABLE_E_DR_CAPABLE - debug request capability (R/O)
 *
 * Value: bits 0..15 = which debug capabilities this bay supports.
 *                     Pack 19 will support them all, pack 16 supports
 *                     none.  The bits are as defined in dfi_debug_request.h
 *        bits 16..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_E_DR_CAPABLE      (130<<3)/* debug request capability (R/O) */

/**
 * CABLE_WE_DR - debug request from driver.  This cable register
 * allows you to write a value to the debug_request variable in the
 * NIOS, or read its current value.  The bit definitions of the
 * debug_request variable are defined in dfi_debug_request.h.  If you
 * set any of the bits via a cable register write, this will initiate
 * some debug requests by the NIOS, which allow you to test or debug
 * the system.
 *
 * Value: bits 0..15 = the debug_request values (see dfi_debug_request.h)
 *        bits 16..63 = unused
 *
 * Notes: introduced in pack 19
 */
#define CABLE_WE_DR             (131<<3)/* debug request from driver */

/*
 * This is the largest cable register.  NOTE: When you add new cable
 * registers to the end of this list, update CABLE_MAX_REGISTER to be
 * the largest cable register.
 */
#define CABLE_MAX_REGISTER      CABLE_WE_DR

#endif /* FIRMWARE_MSW_SRC_DRIVERS_H */
