/*
 * drivers/library/qd_ioctl.h
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#ifndef DRIVERS_LIBRARY_QD_IOCTL_H
#define DRIVERS_LIBRARY_QD_IOCTL_H

/*
 * This is for the definitions of the _IO, _IOR, _IOW and _IOWR macros,
 * used to construct the ioctl numbers.
 *
 * These macros are absolutely essential for some operating systems,
 * and optional for others.  Because we want to be portable across a wide
 * range, these macros shall be used for all operating systems.
 *
 * There is no expectation that the actual *numbers* will be the the
 * same on all operating systems.  They each have their own scheme for
 * encoding argument direction and size.
 */
#ifdef __linux__
  /* This is because Linux gets it's include insulation wrong */
  #define winsize qd_winsize_hack
  #define termio qd_termio_hack
#endif
#ifdef __sun__
  #ifndef BSD_COMP
    #define qd_need_to_undef_BSD_COMP
    #define BSD_COMP
  #endif
#endif
#include <sys/ioctl.h>
#undef winsize
#undef termio
#ifdef qd_need_to_undef_BSD_COMP
  #undef qd_need_to_undef_BSD_COMP
  #define BSD_COMP
#endif

/*
 * This is the architecture-specific set of generated types, so that we
 * can use the same type names on all operating systems, and expect to
 * get the same sized data elements.
 */
#include <qd_type.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
  * The QD_IOCTL_ID symbol is the "high bits" of the QikDATA ioctl
  * numbers.  It is an attempt to make use relatively unique in the
  * ioctl number space.  You can only rely on getting 8 bits, although
  * some operating systems give you more.
  */
#define QD_IOCTL_ID 0x51

/**
  * The QD_IOCTL_HELLO ioctl is used to determine if the disk you are taking
  * to is really a QikDATA.
  *
  * This ioctl is intended to be used as follows
  *	qd_ioctl_hello_t result;
  *	if (ioctl(fd, QD_IOCTL_HELLO, &result))
  *     {
  *	    fatal_wi_errno("ioctl QD_IOCTL_HELLO %s", filename);
  *     }
  *	if (strncmp(result.signature, QD_IOCTL_HELLO_SIGNATURE,
  *	    sizeof(result.signature))
  *	{
  *	    fatal("%s: not a QikDATA", filename);
  *	}
  *
  * It is absolutely VERY IMPORTANT that the size given is sizeof the
  * first argument, othewise the comparison will potentially go off into
  * la-la land, or give false negatives.
  *
  * Within the driver, a strncpy (note the n) shall be used to set the
  * signature string.  Again, the length is the sizeof the first argument.
  */

typedef struct qd_ioctl_hello_t qd_ioctl_hello_t;
struct qd_ioctl_hello_t
{
    /**
      * The signature shall be set to as much of the QD_IOCTL_HELLO_SIGNATURE
      * value as will fit, and padded with zeros.
      */
    char    signature[16];

    /**
      * The instance shall be set to a value meaningful only to the
      * device driver, but shall be unique to each unique discovered
      * QikDATA drive.  For example, it could be set to the device number
      * (major, minor) of the device, with the partiton shifted off,
      * or masked off.
      */
    qd_uint32_t instance;
};

#define QD_IOCTL_HELLO _IOR(QD_IOCTL_ID, 1, qd_ioctl_hello_t)
#define QD_IOCTL_HELLO_SIGNATURE "Platypus QikDATA"

#ifdef __cplusplus
}
#endif

#endif /* DRIVERS_LIBRARY_QD_IOCTL_H */
