/*
 * drivers/linux2.4/qikdata_drive.h
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#ifndef DRIVERS_LINUX2_4_QIKDATA_DRIVE_H
#define DRIVERS_LINUX2_4_QIKDATA_DRIVE_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * If you need to turn on or off debugging, aecp the qikdata_debug.h
 * file and edit the QIKDATA_DEBUG define.  DO NOT edit this file.
 * DO NOT edit the module.cook file.
 */
#include <qikdata_debug.h>

/*
 * If you want the ATTN led to flash during interrupt requests to provide
 * some feedback that the qikData is operating then define FLASHER
 */
#define FLASHER 1

/*
 * This is the long name of the driver, used in our error messages,
 * used in the /proc file system entry, and other places where our short
 * name is not sufficiently informative.
 */
#define DRIVER_LONG_NAME "qikdata"

/*
 * There HERE macro may be used in debug messages, for example
 *
 *      HDEBUG(KERN_DEBUG HERE "Something happened\n");
 *
 * When QIKDATA_DEBUG is defined, it inserts the device long name,
 * the file name and the line number.  When QIKDATA_DEBUG is undefined,
 * it only inserts the device long name.  There is no run-time hit,
 * the compiler does all the work.
 */
#ifdef QIKDATA_DEBUG
#define HERE  DRIVER_LONG_NAME ": " __FILE__ ": " __MODULE_STRING(__LINE__) ": "
#else
#define HERE  DRIVER_LONG_NAME ": "
#endif

#include <linux/ioctl.h>

/* version dependencies have been confined to a separate file */

#include <sysdep.h>

#undef HDEBUG
#ifdef QIKDATA_DEBUG
#  define HDEBUG( fmt, args... ) printk( fmt, ## args )
#else
#  define HDEBUG( fmt, args... )
#endif

#define QD_MAXRDEV   32        /* maximum number of physical devices */
                               /* (HBAs * drives) we'll allow */
#define QIKDRIVE_MAJOR 0       /* dynamic major by default */
#define QIKDRIVE_DEVS 2        /* two disks */
#define QIKDRIVE_RAHEAD 0      /* one sector */
#define QIKDRIVE_BLKSIZE 1024  /* 1k blocks */
#define QIKDRIVE_HARDSECT 512  /* FIXME  are 1k hw sectors doable??? */

/*
 * Split device number into HBA and drive parts.  HBA indexes into the
 * devExt array and drive indexes into the unitDet array.
 *
 *         7     6     5     4     3     2     1     0
 *	+-----+-----+-----+-----+-----+-----+-----+-----+
 *      |    HBA    |      drive      |    partition    |
 *	+-----+-----+-----+-----+-----+-----+-----+-----+
 *
 * If you change these definitions, remmeber to change
 * programs/qd_mknod/os/linux.c as well.
 */

#define MAX_QIK_ADAPS 4
#define MAX_QD_NR     16   /* currently the maximum possible address */

#define MAX_QIK_DATAS 7    /* zeroth qikData is the qikAdaptor */
#define MAX_PARTS 7       /* zeroth partition is whole device */
/*
 * FUTURE: The constants for shifts and masks in these macros should be
 * derived from MAX_QIK_ADAPS, MAX_QIK_DATA and MAX_PARTS.
 */
#define QD_HBA_NR(i_rdev)   ((MINOR(i_rdev) & 0xC0) >> 6)
#define QD_DRV_NR(i_rdev)   ((MINOR(i_rdev) & 0x38) >> 3)
#define QD_MK_MIN(dev,drv,part) ( ((dev&3)<<6) | ((drv&7)<<3) | (part&7) )

#define MAX_CARDS  MAX_QIK_ADAPS

/* for debug print statements */
#define HEX8    "%08X"
#define HEX4    "%04X"
#define HEX2    "%02X"

/*
 *      Some global constants and definitions
 */

#define SDRAM_PAGE_SIZE         4096
#define SDRAM_PAGE_BLOCKS       (SDRAM_PAGE_SIZE/512)
#define MAX_TRANSFER_SIZE       (256*1024)
#define MAX_BANKS               32

enum DEV_STATE
{
    DEV_STATE_INVALID,
    DEV_STATE_DATA,         /* normal data transfer */
    DEV_STATE_SENSEINFO,    /* senseinfo transfer */
    DEV_STATE_INITSDRAM,    /* initialise SDRAM */
    DEV_STATE_FORMATUNIT    /* format unit command */
};
typedef enum DEV_STATE DEV_STATE;

#define MAXIMUM_SGL_DESCRIPTORS     32

struct SG_DESCRIPTOR
{
    u32 PCIAddressLow;
    u32 LocalAddress;
    u32 TransferSize;
    u32 NextDescriptor;
    u32 PCIAddressHigh;
    u32 _filler[3];         /* force eight-word alignment */
};
typedef struct SG_DESCRIPTOR SG_DESCRIPTOR;

struct SGL
{
    SG_DESCRIPTOR   Descriptor[MAXIMUM_SGL_DESCRIPTORS+8];
};
typedef struct SGL SGL;
typedef SGL *PSGL __attribute__ ((aligned(16)));

#define     MAX_DRIVES      8   /* only used to set UnitDet array size */
#define     MAX_DIMMS       8

typedef struct DIMM_PARAMS DIMM_PARAMS;
struct DIMM_PARAMS
{
    u8      manufacturer_jedec_id[8];       /* spd 64-71 */
    u32     assembly_serial;                /* spd 95-98 */

    u8      cols_log2;
    u16     cols_total;
    u16     cols_mask;
    u8      rows_log2;
    u16     rows_total;
    u16     rows_mask;
    u8      banks_log2;
    u8      banks_total;
    u8      banks_mask;
    u8      rsrows_log2;
    u8      rsrows_total;
    u8      rsrows_mask;
    u8      ecc;

    /* mainly for optimal access to geometry */
    u32     startBlock;
    u32     blocks;
    u32     totalBlocks;
    u32     baseAddress;

};
typedef DIMM_PARAMS *PDIMM_PARAMS;

typedef struct DIMM_STATS DIMM_STATS;
struct DIMM_STATS
{
    u32     read_blocks;
    u32     write_blocks;
    u32     uncorrectable;
    u32     correctable;

};
typedef DIMM_STATS *PDIMM_STATS;

#define SIZE_DIMM_STATS (sizeof(DIMM_STATS))
#define SIZE_ALL_DIMM_STATS (SIZE_DIMM_STATS*MAX_DIMMS)
/*
 * note: we're taking into account that the structure contains only ULONGs
 *       and hence will not have any padding bytes embedded in it
 */

/* this needs to be less than 1 page (4096 bytes) in length */
typedef struct NONCACHED_EXTENSION NONCACHED_EXTENSION;
struct NONCACHED_EXTENSION
{
    u8      eccBuffer[SIZE_ALL_DIMM_STATS];  /*  128 bytes */
    u16     bankBuffer[MAX_BANKS];           /*   64 bytes */
    SGL     Sgl;                             /* 1280 bytes */
};
typedef NONCACHED_EXTENSION *PNONCACHED_EXTENSION;

/* 1 page in length */
#define IOBUFFER_SIZE (PAGE_SIZE)
#define IOBUFFER_SECTORS (IOBUFFER_SIZE/512)

/*
 * Note on pointer sizes - use platform defined u_long for
 * pointers BARs etc because all the linux functions want it
 * that way.
 *
 * UNIT_DETAILS holds all the the drive specific details.
 * It is being set up here in preparation for a multi-drive
 * version of the driver.
 */
typedef struct UNIT_DETAILS UNIT_DETAILS;
struct UNIT_DETAILS
{
    char    drive_name[16];
    int     drive_address;
    u_long  drive_type;
    u_long  drive_serial_number;
    u_short ApexVersion;
    u_short MaxVersion;
    u_short MonVersion;
    u_short UPSVersion;
    u_char  SwVersion[8+1];
    u_char  SwBuild[8+1];

    u_int   pack_version; /* holds the pack version for this QikData */
    /*
     * holds the earliest pack version that this QikData
     * can work with
     */
    u_int   earliest_compatible_pack;

    DIMM_PARAMS Dimm[MAX_DIMMS];
    u8      MaxDimmRows;
    u8      MaxDimmCols;
    u8      MaxDimmB;
    u8      ECCInstalled;
    DIMM_STATS DimmStats[MAX_DIMMS];

    u8      UnitReady;

    DEV_STATE State;
};
typedef UNIT_DETAILS *PUNIT_DETAILS;

#define PENDING_REQUEST_BOGUS 0XFF

typedef struct HW_DEVICE_EXTENSION HW_DEVICE_EXTENSION;
struct HW_DEVICE_EXTENSION
{
    char    qd_name[16];
    int     major;
    unsigned char irq;
    unsigned char pci_bus;
    unsigned char pci_device_fn;
    struct pci_dev *pci_dev;
    spinlock_t lock;

    int     DeviceIndex;
    int     devAddr;
    int     nBays;
    int     UnitIndex;
    int     CurrUnit;

    /* need date as well    StartDate; */
    struct timeval StartTime;

    u_long  BAR[4];
    u32     BARrange[4];
    u8      BARtype[4];
    void    *LASmap;
    void    *RTRmap;

    PNONCACHED_EXTENSION NoncachedExtension;
    void    *ioBuffer;

    u_long  eccBufferAddr;
    u_long  bankBufferAddr;
    u_long  ioBufferAddr;

    u32     AcexIDReg;
    u32     AcexIOReg;
    PAGE_REG AcexPageReg;

    u16     QikDriveType;
    u32     SerialNumber;
    u8      MaxVersion;
    u16     FlexStatCntl;

    u32     SdramInitBlitBlocks;
    u32     FormatProgress;
    u8      DeviceReady;
    DEV_STATE State;
    u8      PendingRequest;

    /* Driver statistics/diagnostics */

    u_long  LargestTransferSize;
    char    ReceivedUnaligned;
    int     usage;

    /*
     * the array UnitDet holds details for each QikDATA drive attached
     * to this HBA.
     */
    UNIT_DETAILS UnitDet[MAX_DRIVES];
};
typedef HW_DEVICE_EXTENSION *PHW_DEVICE_EXTENSION;

extern HW_DEVICE_EXTENSION qikdata_device_extension[];

#endif /* DRIVERS_LINUX2_4_QIKDATA_DRIVE_H */
