
/*
 *  Broadcom Cryptonet Driver software is distributed as is, without any warranty
 *  of any kind, either express or implied as further specified in the GNU Public
 *  License. This software may be used and distributed according to the terms of
 *  the GNU Public License.
 *
 * Cryptonet is a registered trademark of Broadcom Corporation.
 */
/******************************************************************************
 *
 *  Copyright 2000
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/

/*******************************************************************************
 *
 * File: Linux/tlsmac.c
 * 
 * What: Character driver interface to TLS HMAC routines in the ubsec driver
 *
 * Revision History:
 *                   When       Who   What
 *                   10/24/00   DNA   Created
 * March 2001 PW Release for Linux 2.4 UP and SMP kernel 
 *
 ******************************************************************************/

#include "cdevincl.h"

/* This is useful only for diagnostics. */
#undef STATIC_ALLOC_OF_CRYPTO_BUFFERS

#ifdef POLL
#undef GOTOSLEEP
#endif

#ifdef STATIC_ALLOC_OF_CRYPTO_BUFFERS

unsigned char * pKernSourceBuffer = NULL;
unsigned char * pKernDestBuffer   = NULL;
static int      PhysSourceBuf     = 0;
static int      PhysDestBuf       = 0;

#endif /* ifdef STATIC_ALLOC_OF_CRYPTO_BUFFERS */


/**************************************************************************
 *
 * Function:  init_tlsmacif
 * 
 * Called from: init_module() in Linux/dispatch.c
 *
 * Description: 
 *
 * Return Values:
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int 
init_tlsmacif(void) {
  int error = 0;
  
#ifdef STATIC_ALLOC_OF_CRYPTO_BUFFERS
  
  pKernSourceBuffer = pKernDestBuffer = (char *)kmalloc((MAX_FILE_SIZE),GFP_KERNEL|GFP_ATOMIC);
  if(pKernSourceBuffer == NULL) {
    printk("Cryptonet: no memory for source buffer %d\n", MAX_FILE_SIZE);
    error = -ENOMEM;
    return(error);
  }
  memset(pKernSourceBuffer,0, MAX_FILE_SIZE);
  
#ifdef DEBUG
  
  printk("Allocate Source: %x %x\n", pKernSourceBuffer, vtophys(pKernSourceBuffer));

#endif /* ifdef DEBUG */
  
  PhysSourceBuf = PhysDestBuf = virt_to_bus(pKernSourceBuffer);

#ifdef DEBUG
  
  printk("Cryptonet: Memory Alloc Source %x %x Dest %x %x for source buffer\n",
	 pKernSourceBuffer, PhysSourceBuf, KernDestBuffer, PhysDestBuf);
  printk("Cryptonet: Memory Alloc State %x %x state buffer\n", pKernStateBuffer, PhysStateBuf);
    
#endif /* ifdef DEBUG */
  
#endif /* ifdef STATIC_ALLOC_OF_CRYPTO_BUFFERS */

  return(error);
}


/**************************************************************************
 *
 * Function: shutdown_tlsmacif
 *
 * Called from: cleaup_module() in Linux/dispatch.c
 *
 * Description:
 *
 *************************************************************************/

void 
shutdown_tlsmacif(void) {
  
#ifdef STATIC_ALLOC_OF_CRYPTO_BUFFERS
  
  if(pKernSourceBuffer != NULL)
    kfree(pKernSourceBuffer);
  
  pKernSourceBuffer = pKernDestBuffer = NULL;
  PhysSourceBuf = PhysDestBuf = 0;
  
#endif /* ifndef STATIC_ALLOC_OF_CRYPTO_BUFFERS */
  
  return;
}

/**************************************************************************
 *
 * Function:  ubsec_tlsmac
 * 
 * Called from: ubsec_ioctl() in Linux/dispatch.c
 *
 * Purpose:
 *          Wrapper function between user call and call to SRL.
 *
 * Description: 
 *              Prepare all the data from the user's call for a call to 
 *              the SRL, set up the timer, and finally place this thread
 *              on the wait queue and go to sleep. When ubsec_tlsmac is 
 *              called, the relevant data for an SSL DES op reside in
 *              user space. The struct ubecom_tlsmac_io_pt argument, pIOInfo, 
 *              points to data on the stack and in kernal space. However, 
 *              pIOInfo is a struct that contains pointers which still 
 *              point to memory in user space. ubecom_ssl copies the 
 *              contents of pIOInfo as well as the input data it points
 *              to into memory allocated in kernal space with LinuxAllocateMemory.
 *
 * Return Values: 
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int
ubsec_tlsmac(
		ubsec_DeviceContext_t  pContext,
		ubsec_tlsmac_io_pt *   ppIOInfo
		) {
  
  ubsec_tlsmac_io_t             IOInfo;
  ubsec_tlsmac_io_pt            pIOInfo           = &IOInfo;
  ubsec_SSLCommandInfo_t        SslCommand;
  ubsec_SSLCommandInfo_pt       pSslCommand       = &SslCommand;
  ubsec_TLSHMACParams_pt        pTLSHMACParams     = &(pSslCommand->Parameters.TLSHMACParams);
  volatile CommandContext_t  CommandContext;
  CommandContext_t          *pCommandContext   = &CommandContext;
  int                           NumCommands       = 1;
  unsigned long		        delayTotalUs      = 0;
  ubsec_FragmentInfo_pt         pSourceFragments  = NULL;
  unsigned char *               pUserSourceBuffer = NULL;
  unsigned int                  SourceBufferBytes = 0;
  unsigned int                  DestBufferBytes   = UBSEC_HMAC_LENGTH;
  int                           error             = 0;
  int                           i                 = 0;
  ubsec_HMAC_State_t            HMACState;
  ubsec_HMAC_State_pt           pHMACState        = &HMACState;
  
#ifndef STATIC_ALLOC_OF_CRYPTO_BUFFERS
  
  unsigned char *              pKernSourceBuffer = NULL;
  unsigned char *              pKernDestBuffer   = NULL;
  int                          PhysSourceBuf     = 0;
  int                          PhysDestBuf       = 0;
  
#endif /* ifndef STATIC_ALLOC_OF_CRYPTO_BUFFERS */

  copy_from_user(pIOInfo, *ppIOInfo, sizeof(ubsec_tlsmac_io_t));
  pUserSourceBuffer = pIOInfo->SourceBuffer;
  SourceBufferBytes = pIOInfo->SourceBufferBytes;

#ifdef STATIC_ALLOC_OF_CRYPTO_BUFFERS
  
  if(SourceBufferBytes > MAX_FILE_SIZE) {
    printk("Cryptonet: input file too large <%d,%d>\n", SourceBufferBytes, MAX_FILE_SIZE);
    error = EINVAL;
    return(error);
  }

#else
  
  pKernSourceBuffer = (char *)kmalloc((SourceBufferBytes),GFP_KERNEL|GFP_ATOMIC);
  if(pKernSourceBuffer == NULL) {
    printk("Cryptonet: no memory for source buffer %d\n", SourceBufferBytes);
    error = -ENOMEM;
    goto ReturnErrorLabel;
  }
  memset(pKernSourceBuffer,0, SourceBufferBytes);
  
  pKernDestBuffer = (char *)kmalloc((DestBufferBytes),GFP_KERNEL|GFP_ATOMIC);
  if(pKernDestBuffer == NULL) {
    printk("Cryptonet: no memory for dest buffer %d\n", DestBufferBytes);
    error = -ENOMEM;
    goto ReturnErrorLabel;
  }
  memset(pKernDestBuffer,0, DestBufferBytes);
  
#ifdef DEBUG
  printk("Allocate Source: %x %x\n", pKernSourceBuffer, vtophys(pKernSourceBuffer));
  printk("Allocate Dest: %x %x\n", pKernDestBuffer, vtophys(pKernDestBuffer));
#endif
  
  PhysSourceBuf = virt_to_bus(pKernSourceBuffer);
  PhysDestBuf   = virt_to_bus(pKernDestBuffer);

#ifdef DEBUG
  printk("Cryptonet: Memory Alloc Source %x %x Dest %x %x for source buffer\n",
	 pKernSourceBuffer, PhysSourceBuf, pKernDestBuffer, PhysDestBuf);
#endif
  
#endif /* ifdef STATIC_ALLOC_OF_CRYPTO_BUFFERS */
  
  memcpy(pHMACState->InnerState, pIOInfo->InnerState, UBSEC_HMAC_LENGTH);
  memcpy(pHMACState->OuterState, pIOInfo->OuterState, UBSEC_HMAC_LENGTH);
  
  pTLSHMACParams->HMACState      = pHMACState;
  pTLSHMACParams->Version        = pIOInfo->Version;
  pTLSHMACParams->SequenceNumber = pIOInfo->SequenceNumber;
  pTLSHMACParams->ContentType    = pIOInfo->ContentType;
  pTLSHMACParams->DataLength     = pIOInfo->DataLength;
  
#if DEBUG
  printk("hmacstate = %x\n", pTLSHMACParams->HMACState);
#endif

  /* Assemble Source Fragments */
  
  pSourceFragments = kmalloc((((sizeof(ubsec_FragmentInfo_t) * UBSEC_MAX_FRAGMENTS))),GFP_KERNEL|GFP_ATOMIC);
  if(pSourceFragments == NULL) {
    printk("Cryptonet: no memory for fragment buffer\n");
    error = -ENOMEM;
    return(error);
  }
  
  pSslCommand->SourceFragments = pSourceFragments;
  pSslCommand->NumSource       = SetupFragmentList(pSslCommand->SourceFragments,
						   pKernSourceBuffer,
						   SourceBufferBytes);
  
  copy_from_user(pKernSourceBuffer, pUserSourceBuffer, SourceBufferBytes);
  
  /* Assemble Destination Fragments */

  pSslCommand->NumDestination               = 0;
  pTLSHMACParams->OutputHMAC.FragmentAddress = PhysDestBuf;
  pTLSHMACParams->OutputHMAC.FragmentLength  = 0; /* fixed output length */
  
  pCommandContext->CallBackStatus   = 0;
  pSslCommand->CommandContext       = pCommandContext;
  pSslCommand->CompletionCallback   = CmdCompleteCallback;
  pSslCommand->Command              = pIOInfo->command;
  
  start_time(&(pCommandContext->tv_start));

#if DEBUG
  printk("Linux:ubsec_sslarc4 just before SRL call...\n");
#endif

#ifndef LINUX2dot2
  init_waitqueue_head(&pCommandContext->WaitQ);
#else
   pCommandContext->WaitQ         = 0;
#endif 

  pIOInfo->result_status = ubsec_SSLCommand(pContext, pSslCommand, &NumCommands);
  
  switch(pIOInfo->result_status) {
  case UBSEC_STATUS_SUCCESS:
    break;
    
  case UBSEC_STATUS_TIMEOUT:
    printk("ubsec error: ubsec_SslCommand() timeout\n");
    ubsec_ResetDevice(pContext);
    error = ETIMEDOUT;
    goto ReturnErrorLabel;
    break;
    
  case UBSEC_STATUS_INVALID_PARAMETER:
    printk("ubsec error:  ubsec_SslCommand() invalid parameter\n");
    error = EINVAL;
    goto ReturnErrorLabel;
    break;
    
  case UBSEC_STATUS_NO_RESOURCE:
    printk("ubsec error: ubsec_SslCommand() no resource. Number done: %d\n", NumCommands);
    error = ENOBUFS;
    goto ReturnErrorLabel;
    break;
    
  default:
    error = ENOMSG;
    goto ReturnErrorLabel;
    break;
  }
  
#ifdef GOTOSLEEP
  
  if(!(CommandContext.CallBackStatus))  /* Just in case completed on same thread. */
    Gotosleep(&CommandContext.WaitQ);
    if (!pCommandContext->CallBackStatus) {
              pCommandContext->Status=UBSEC_STATUS_TIMEOUT;
    }                          
  
#else
  
  for(delayTotalUs = 1; !(CommandContext.CallBackStatus); delayTotalUs++) {
    
#ifdef POLL
    
    /* We need to poll the device if we are operating in POLL mode. */
    ubsec_PollDevice(pContext);
    
#endif
    
    if(delayTotalUs >= 3000000) {
      error = ETIMEDOUT;
      goto ReturnErrorLabel;
    }
    udelay(1);
  }
  
#endif

#if DEBUG
  printk("Linux: Dest Buffer (post-SRL)\n");
  for(i = 0; i < DestBufferBytes; i++) {
    printk("%02X", pKernDestBuffer[i]);
  }
  printk("\n");
#endif
  
  pIOInfo->time_us = CommandContext.tv_start.tv_sec * 1000000 + CommandContext.tv_start.tv_usec;
  if(pIOInfo->result_status == UBSEC_STATUS_SUCCESS) {
    copy_to_user(pIOInfo->HmacBuffer, pKernDestBuffer, DestBufferBytes);
    copy_to_user(*ppIOInfo, pIOInfo, sizeof(ubsec_arc4_io_t));
  }
  
 ReturnErrorLabel:
  
#ifndef STATIC_ALLOC_OF_CRYPTO_BUFFERS
  
  if(pKernSourceBuffer != NULL)
    kfree(pKernSourceBuffer);
  if(pKernDestBuffer != NULL)
    kfree(pKernDestBuffer);
  pKernSourceBuffer = pKernDestBuffer = NULL;
  PhysSourceBuf = PhysDestBuf = 0;
  
#endif /* ifndef STATIC_ALLOC_OF_CRYPTO_BUFFERS */
  
  if(pSourceFragments != NULL) {
    kfree(pSourceFragments);
    pSourceFragments = NULL;
  }

  return(error);
}

