#ifndef __LINUX_HOOK_H
#define __LINUX_HOOK_H
/*****************************************************************************/
/* Kernel Hooks Interface.                                    		     */
/* Version 1.2                                                               */
/* Author: Richard J Moore richardj_moore@uk.ibm.com                         */
/* 	   Vamsi Krishna S. r1vamsi@in.ibm.com 				     */
/*                                                                           */
/* Kernel hook interface definitions and arch-independent hooks.	     */
/*                                                                           */
/* Refer to Documentation/hook/HOWTO for details.                            */
/*****************************************************************************/
#include <linux/compiler.h>

/*
 * Use the DECLARE_HOOK_HEAD macro to define the hookhead structure in global 
 * memory of a kernel module that is implementing a hook.
 */
#if defined(CONFIG_HOOK) || defined(CONFIG_HOOK_MODULE)
#define DECLARE_HOOK_HEAD(x) _DECLARE_HOOK_HEAD(x)
#define _DECLARE_HOOK_HEAD(hk) \
char hk ## _id[sizeof(#hk)] = #hk; \
extern void hk; \
hook_head_t hk ## _head = { \
	hook_head: &(hk##_head), \
	hook_loc: &(hk), \
	hook_id: hk##_id \
}; \
EXPORT_SYMBOL(hk ## _head); \
extern hook_head_t hk ## _head

#define DECLARE_EXCLUSIVE_HOOK_HEAD(x) _DECLARE_EXCLUSIVE_HOOK_HEAD(x)
#define _DECLARE_EXCLUSIVE_HOOK_HEAD(hk) \
char hk ## _id[sizeof(#hk)] = #hk; \
extern void hk; \
hook_head_t hk ## _head = { \
	hook_head: &(hk##_head), \
	hook_loc: &(hk), \
	hook_id: hk##_id, \
	hook_flags: HOOK_HEAD_EXCLUSIVE \
}; \
EXPORT_SYMBOL_NOVERS(hk ## _head); \
extern hook_head_t hk ## _head

#define USE_HOOK(x) _USE_HOOK(x)
#define _USE_HOOK(hk) extern hook_head_t hk ## _head
#else
#define DECLARE_HOOK_HEAD(x)
#define USE_HOOK(x)
#endif

/* define the user (kernel module) hook structure */
struct hook_rec_s;
struct hook_head_s;

typedef struct hook_rec_s {
	void * hook_exit;
	char * hook_exit_name;
	struct hook_rec_s * hook_next;
	struct hook_rec_s * hook_prev;
	unsigned int hook_flags;
	struct hook_rec_s * hook_link;
	struct hook_head_s * hook_head;
	unsigned int hook_index;
	struct proc_dir_entry *proc_entry;
	int proc_writable;
} hook_rec_t;

typedef struct hook_head_s {
	void * hook_exit;
	char * hook_exit_name;
	struct hook_rec_s * hook_next;
	struct hook_rec_s * hook_prev;
	unsigned int hook_flags;
	struct hook_rec_s * hook_link;
	struct hook_head_s * hook_head;
	unsigned int hook_index;
	struct proc_dir_entry *proc_entry;
	struct hook_head_s * hook_head_next;
	void * hook_loc;
	char * hook_id;
} hook_head_t;

/* hook flags */
#define HOOK_ARMED		0x00000001
#define HOOK_REGISTERED		0x00000002
#define HOOK_PRIORITY_FIRST	0x00000004
#define HOOK_PRIORITY_LAST	0x00000008
#define HOOK_PRIORITY_EXCLUSIVE	0x0000000c
#define HOOK_QUEUE_LIFO		0x00000010

/* the following are reserved for future enhancement */
#define HOOK_OPTION_DISABLE_IRQS	0x00001000
#define HOOK_OPTION_STOP_CPUS		0x00002000


/* flag groupings */
#define HOOK_STATUS		(HOOK_ARMED | HOOK_REGISTERED)
#define HOOK_PRIORITY		(HOOK_PRIORITY_FIRST | HOOK_PRIORITY_LAST)
#define HOOK_QUEUE		HOOK_QUEUE_LIFO
#define HOOK_OPTIONS		(HOOK_OPTION_DISABLE_IRQS | HOOK_OPTION_STOP_CPUS)
#define HOOK_VALID_USER_FLAGS	(HOOK_PRIORITY | HOOK_QUEUE | HOOK_OPTIONS)


/* exported functions wrapped in a macro */
int _hook_initialise(hook_head_t *);
int _hook_terminate(hook_head_t *, int);
int _hook_exit_register(char *, hook_rec_t *, int proc_writable);

/* exported functions error codes */
#define NO_ERROR 				 0
#define ERROR_HOOK_NOT_FOUND 			-1
#define ERROR_HIGHER_PRIORITY_HOOK 		-2
#define ERROR_INVALID_REGISTRATION_RECORD 	-3
#define ERROR_LOWER_PRIORITY_HOOK 		-4
#define ERROR_GKHI_UNAVAILABLE 			-5
#define ERROR_INVALID_HOOK_LOCATION 		-6
#define ERROR_HOOK_EXISTS 			-7
#define ERROR_HOOK_IS_ACTIVE 			-8

/* Return values from Hook Exit routines */
#define HOOK_CONTINUE 	0
#define HOOK_TERMINATE 	1
#define HOOK_RETURN 	-1

/* exported function prototypes */
int hook_exit_deregister(hook_rec_t *);
int hook_exit_arm(hook_rec_t *);
int hook_exit_disarm(hook_rec_t *);

/*
 * public interfaces requiring a macro wrapper
 *
 * int hook_exit_register(hookid, hook_rec_t *);
 * int hook_terminate(hookid, int force_flag);/
 * int hook_initialise(hookid);
 *
 * where hookid is a typeless name
 */
#define hook_exit_register(x, y, z) dummy_hook_exit_register(x, y, z)
#define dummy_hook_exit_register(x, y, z) _hook_exit_register(#x, y, z)
#define hook_terminate(x, y) dummy_hook_terminate(x, y)
#define dummy_hook_terminate(x, y) _hook_terminate(&(x ## _head), y)
#define hook_initialise(x) dummy_hook_initialise(x)
#define dummy_hook_initialise(x) _hook_initialise(&(x ## _head))

/* define head record only flags */
#define HOOK_HEAD_ACTIVE	0x80000000
#define HOOK_HEAD_ASM_HOOK	0x40000000
#define HOOK_HEAD_EXCLUSIVE	0x20000000
#define HOOK_HEAD_VALID_FLAGS	(HOOK_VALID_USER_FLAGS | HOOK_STATUS | HOOK_HEAD_ACTIVE | HOOK_HEAD_ASM_HOOK | HOOK_HEAD_EXCLUSIVE)

/* global status flags */
#define HOOK_INIT       0x00000001
#define HOOK_IN_START   0x00000002


typedef int (*hook_fn_t)(hook_rec_t *, ...);
#include <asm/hook.h>

#if defined(CONFIG_HOOK) || defined(CONFIG_HOOK_MODULE)
#ifndef __HAVE_ASM_HOOK
#define HOOK(h, args...) _HOOK(h, ##args)
#define _HOOK(h, args...) __HOOK(#h, h, ##args)
#define __HOOK(h, h1, args...) \
	__asm__ __volatile__ (".global "h"; "h":"); \
	if (unlikely(h1##_head.hook_flags & HOOK_HEAD_ACTIVE)) { \
		register hook_rec_t *ptr; \
		extern hook_head_t h1##_head; \
		for (ptr = h1##_head.hook_next; \
			!((ptr->hook_flags & HOOK_ARMED) && \
			  ((*(hook_fn_t)(ptr->hook_exit))(ptr, ##args))); \
			ptr = ptr->hook_next); \
	}
#endif /* __HAVE_ASM_HOOK */

#ifndef __HAVE_ASM_HOOK_RET
#define HOOK_RET(h, args...) _HOOK_RET(h, ##args)
#define _HOOK_RET(h, args...) __HOOK_RET(#h, h, ##args)
#define __HOOK_RET(h, h1, args...) \
	__asm__ __volatile__ (".global "h"; "h":"); \
	if (unlikely(h1##_head.hook_flags & HOOK_HEAD_ACTIVE)) { \
		typedef int (*f)(hook_rec_t *, int *, ...); \
		register hook_rec_t *ptr; \
		int rc; \
		register int dsprc; \
		extern hook_head_t h1##_head; \
		for (ptr = h1##_head.hook_next; \
			!((ptr->hook_flags & HOOK_ARMED) && \
			  (dsprc=(*(hook_fn_t)(ptr->hook_exit))(ptr, &rc, ##args))); \
			ptr=ptr->hook_next); \
		if (dsprc == HOOK_RETURN) return rc; \
	}
#endif /* __HAVE_ASM_HOOK_RET */

#ifndef __HAVE_ASM_HOOK_RET_NORC
#define HOOK_RET_NORC(h, args...) _HOOK_RET_NORC(h, ##args)
#define _HOOK_RET_NORC(h, args...) __HOOK_RET_NORC(#h, h, ##args)
#define __HOOK_RET_NORC(h, h1, args...) \
	__asm__ __volatile__ (".global "h"; "h":"); \
	if (unlikely(h1##_head.hook_flags & HOOK_HEAD_ACTIVE)) { \
		typedef int (*f)(hook_rec_t *, ...); \
		register hook_rec_t *ptr; \
		register int dsprc; \
		extern hook_head_t h1##_head; \
		for (ptr = h1##_head.hook_next; \
			!((ptr->hook_flags & HOOK_ARMED) && \
			  (dsprc=(*(hook_fn_t)(ptr->hook_exit))(ptr, ##args))); \
			ptr = ptr->hook_next); \
	 	if (dsprc == HOOK_RETURN) return; \
	}
#endif /* __HAVE_ASM_HOOK_RET_NORC */

#ifndef __HAVE_ASM_EXCLUSIVE_HOOK
#define EXCLUSIVE_HOOK(h, args...) _EXCLUSIVE_HOOK(h, ##args)
#define _EXCLUSIVE_HOOK(h, args...) __EXCLUSIVE_HOOK(#h, h, ##args)
#define __EXCLUSIVE_HOOK(h, h1, args...) \
	__asm__ __volatile__ (".global "h"; "h":"); \
	if (unlikely(h1##_head.hook_flags & HOOK_HEAD_ACTIVE)) { \
		typedef int (*f)(hook_rec_t *, ...); \
		extern hook_head_t h1##_head; \
		(*(hook_fn_t)(h1##_head.hook_exit))(h1##_head.hook_next, ##args); \
	}
#endif /* __EXCLUSIVE_HAVE_ASM_HOOK*/

#ifndef __HAVE_ASM_EXCLUSIVE_HOOK_RET
#define EXCLUSIVE_HOOK_RET(h, args...) _EXCLUSIVE_HOOK_RET(h, ##args)
#define _EXCLUSIVE_HOOK_RET(h, args...) __EXCLUSIVE_HOOK_RET(#h, h, ##args)
#define __EXCLUSIVE_HOOK_RET(h, h1, args...) \
	__asm__ __volatile__ (".global "h"; "h":"); \
	if (unlikely(h1##_head.hook_flags & HOOK_HEAD_ACTIVE)) { \
		typedef int (*f)(int *); \
		int rc; \
		register int dsprc; \
		extern hook_head_t h1##_head; \
		dsprc = (*(hook_fn_t)(h1##_head.hook_exit))(h1##_head.hook_next, &rc, ##args); \
		if (dsprc == HOOK_RETURN) return rc; \
	}
#endif /* __EXCLUSIVE_HAVE_ASM_HOOK_RET */

#ifndef __HAVE_ASM_EXCLUSIVE_HOOK_RET_NORC
#define EXCLUSIVE_HOOK_RET_NORC(h, args...) _EXCLUSIVE_HOOK_RET_NORC(h, ##args)
#define _EXCLUSIVE_HOOK_RET_NORC(h, args...) __EXCLUSIVE_HOOK_RET_NORC(#h, h, ##args)
#define __EXCLUSIVE_HOOK_RET_NORC(h, h1, args...) \
	__asm__ __volatile__ (".global "h"; "h":"); \
	if (unlikely(h1##_head.hook_flags & HOOK_HEAD_ACTIVE)) { \
		typedef int (*f)(void *); \
		extern hook_head_t h1##_head; \
		register int dsprc = (*(hook_fn_t)(h1##_head.hook_exit))(h1##_head.hook_next, ##args); \
		if (dsprc == HOOK_RETURN) return; \
	}
#endif /* __HAVE_ASM_EXCLUSIVE_HOOK_RET_NORC */

/*
 *  Generic hooks are the same in all architectures and may be used to
 *  place hooks even in inline functions. They don't define a symbol at hook
 *  location.
 */ 
#define DECLARE_GENERIC_HOOK_HEAD(x) _DECLARE_GENERIC_HOOK_HEAD(x)
#define _DECLARE_GENERIC_HOOK_HEAD(hk) \
char hk ## _id[sizeof(#hk)] = #hk; \
hook_head_t hk ## _head = { \
	hook_head: &(hk##_head), \
	hook_loc: NULL, \
	hook_id: hk##_id \
}; \
EXPORT_SYMBOL(hk ## _head); \
extern hook_head_t hk ## _head

#define GENERIC_HOOK(h, args...) _GENERIC_HOOK(h, ##args)
#define _GENERIC_HOOK(h, args...) __GENERIC_HOOK(#h, h, ##args)
#define __GENERIC_HOOK(h, h1, args...) { \
	extern hook_head_t h1##_head; \
	if (unlikely(h1##_head.hook_flags & HOOK_HEAD_ACTIVE)) { \
		register hook_rec_t *ptr; \
		for (ptr = h1##_head.hook_next; \
			!((ptr->hook_flags & HOOK_ARMED) && \
			  ((*(hook_fn_t)(ptr->hook_exit))(ptr, ##args))); \
			ptr = ptr->hook_next); \
	} \
}
#else
/* dummy macros when hooks are not compiled in */
#define HOOK(h, args...)
#define HOOK_RET(h, args...)
#define HOOK_RET_NORC(h, args...)
#define EXCLUSIVE_HOOK(h, args...)
#define EXCLUSIVE_HOOK_RET(h, args...)
#define EXCLUSIVE_HOOK_RET_NORC(h, args...)
#endif

#endif /* __LINUX_HOOK_H */
